"""Unit tests for kgb.ops."""

from __future__ import unicode_literals

import re

from kgb.errors import UnexpectedCallError
from kgb.ops import (SpyOpMatchAny,
                     SpyOpMatchInOrder,
                     SpyOpRaise,
                     SpyOpRaiseInOrder,
                     SpyOpReturn,
                     SpyOpReturnInOrder)
from kgb.tests.base import MathClass, TestCase


class SpyOpMatchAnyTests(TestCase):
    """Unit tests for kgb.ops.SpyOpMatchAny."""

    def test_setup_with_instance(self):
        """Testing SpyOpMatchAny set up with op=SpyOpMatchAny([...])"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 1,
                        'b': 2,
                    },
                    'call_fake': lambda a, b: a - b,
                },
            ]))

        self.assertEqual(obj.do_math(a=1, b=2), -1)

    def test_setup_with_instance_and_op(self):
        """Testing SpyOpMatchAny set up with op=SpyOpMatchAny([...]) and op"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 1,
                        'b': 2,
                    },
                    'op': SpyOpMatchInOrder([
                        {
                            'kwargs': {
                                'a': 1,
                                'b': 2,
                                'x': 1,
                            },
                            'op': SpyOpReturn(123),
                        },
                        {
                            'kwargs': {
                                'a': 1,
                                'b': 2,
                                'x': 2,
                            },
                            'op': SpyOpReturn(456),
                        },
                    ]),
                },
            ]))

        self.assertEqual(obj.do_math(a=1, b=2, x=1), 123)
        self.assertEqual(obj.do_math(a=1, b=2, x=2), 456)

    def test_with_function(self):
        """Testing SpyOpMatchAny with function"""
        def do_math(a, b):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpMatchAny([
                {
                    'args': [5, 3],
                    'call_fake': lambda a, b: a - b
                },
            ]))

        self.assertEqual(do_math(5, 3), 2)

    def test_with_function_and_op(self):
        """Testing SpyOpMatchAny with function and op"""
        def do_math(a, b, x=0):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpMatchAny([
                {
                    'args': [5, 3],
                    'op': SpyOpMatchInOrder([
                        {
                            'args': [5, 3],
                            'kwargs': {
                                'x': 1,
                            },
                            'op': SpyOpReturn(123),
                        },
                        {
                            'args': [5, 3],
                            'kwargs': {
                                'x': 2,
                            },
                            'op': SpyOpReturn(456),
                        },
                    ]),
                },
            ]))

        self.assertEqual(do_math(a=5, b=3, x=1), 123)
        self.assertEqual(do_math(a=5, b=3, x=2), 456)

    def test_with_classmethod(self):
        """Testing SpyOpMatchAny with classmethod"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 5,
                        'b': 3,
                    },
                    'call_fake': lambda a, b: a - b
                },
            ]))

        self.assertEqual(MathClass.class_do_math(a=5, b=3), 2)

    def test_with_classmethod_and_op(self):
        """Testing SpyOpMatchAny with classmethod and op"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 5,
                        'b': 3,
                    },
                    'op': SpyOpMatchInOrder([
                        {
                            'kwargs': {
                                'a': 5,
                                'b': 3,
                                'x': 1,
                            },
                            'op': SpyOpReturn(123),
                        },
                        {
                            'kwargs': {
                                'a': 5,
                                'b': 3,
                                'x': 2,
                            },
                            'op': SpyOpReturn(456),
                        },
                    ]),
                },
            ]))

        self.assertEqual(MathClass.class_do_math(a=5, b=3, x=1), 123)
        self.assertEqual(MathClass.class_do_math(a=5, b=3, x=2), 456)

    def test_with_unbound_method(self):
        """Testing SpyOpMatchAny with unbound method"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 3,
                    },
                },
            ]))

        obj = MathClass()
        self.assertEqual(obj.do_math(a=4, b=3), 7)

    def test_with_unbound_method_and_op(self):
        """Testing SpyOpMatchAny with unbound method and op"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 3,
                    },
                    'op': SpyOpMatchInOrder([
                        {
                            'kwargs': {
                                'a': 4,
                                'b': 3,
                                'x': 1,
                            },
                            'op': SpyOpReturn(123),
                        },
                        {
                            'kwargs': {
                                'a': 4,
                                'b': 3,
                                'x': 2,
                            },
                            'op': SpyOpReturn(456),
                        },
                    ]),
                },
            ]))

        obj = MathClass()
        self.assertEqual(obj.do_math(a=4, b=3, x=1), 123)
        self.assertEqual(obj.do_math(a=4, b=3, x=2), 456)

    def test_with_expected_calls(self):
        """Testing SpyOpMatchAny with all expected calls"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 7,
                    },
                },
                {
                    'kwargs': {
                        'a': 2,
                        'b': 8,
                    },
                    'call_original': False,
                },
                {
                    'kwargs': {
                        'a': 100,
                        'b': 200,
                    },
                    'op': SpyOpMatchInOrder([
                        {
                            'kwargs': {
                                'a': 100,
                                'b': 200,
                                'x': 1,
                            },
                            'op': SpyOpReturn(123),
                        },
                        {
                            'kwargs': {
                                'a': 100,
                                'b': 200,
                                'x': 2,
                            },
                            'op': SpyOpReturn(456),
                        },
                    ]),
                },
                {
                    'kwargs': {
                        'a': 5,
                        'b': 9,
                    },
                    'call_fake': lambda a, b: a + b + 10,
                },
                {
                    'a': 2,
                    'call_fake': lambda a, b: 1001,
                },
            ]))

        values = [
            obj.do_math(5, b=9),
            obj.do_math(a=2, b=8),
            obj.do_math(a=100, b=200, x=1),
            obj.do_math(a=100, b=200, x=2),
            obj.do_math(a=1, b=1),
            obj.do_math(4, 7),
        ]

        self.assertEqual(values, [24, None, 123, 456, 1001, 11])

    def test_with_unexpected_call(self):
        """Testing SpyOpMatchAny with unexpected call"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchAny([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 7,
                    },
                },
            ]))

        expected_message = re.escape(
            'do_math was not called with any expected arguments.'
        )

        with self.assertRaisesRegex(AssertionError, expected_message):
            obj.do_math(a=4, b=9)


class SpyOpMatchInOrderTests(TestCase):
    """Unit tests for kgb.ops.SpyOpMatchInOrder."""

    def test_setup_with_instance(self):
        """Testing SpyOpMatchInOrder set up with op=SpyOpMatchInOrder([...])"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 1,
                        'b': 2,
                    },
                },
            ]))

        self.assertEqual(obj.do_math(a=1, b=2), 3)

    def test_setup_with_instance_and_op(self):
        """Testing SpyOpMatchInOrder set up with op=SpyOpMatchInOrder([...])
        and op
        """
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 1,
                        'b': 2,
                    },
                    'op': SpyOpReturn(123),
                },
            ]))

        self.assertEqual(obj.do_math(a=1, b=2), 123)

    def test_with_function(self):
        """Testing SpyOpMatchInOrder with function"""
        def do_math(a, b):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpMatchInOrder([
                {
                    'args': [5, 3],
                    'call_fake': lambda a, b: a - b
                },
            ]))

        self.assertEqual(do_math(5, 3), 2)

    def test_with_function_and_op(self):
        """Testing SpyOpMatchInOrder with function and op"""
        def do_math(a, b):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpMatchInOrder([
                {
                    'args': [5, 3],
                    'op': SpyOpReturn(123),
                },
            ]))

        self.assertEqual(do_math(5, 3), 123)

    def test_with_classmethod(self):
        """Testing SpyOpMatchInOrder with classmethod"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 5,
                        'b': 3,
                    },
                    'call_fake': lambda a, b: a - b
                },
            ]))

        self.assertEqual(MathClass.class_do_math(a=5, b=3), 2)

    def test_with_classmethod_and_op(self):
        """Testing SpyOpMatchInOrder with classmethod and op"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 5,
                        'b': 3,
                    },
                    'op': SpyOpReturn(123),
                },
            ]))

        self.assertEqual(MathClass.class_do_math(a=5, b=3), 123)

    def test_with_unbound_method(self):
        """Testing SpyOpMatchInOrder with unbound method"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 3,
                    },
                },
            ]))

        obj = MathClass()

        self.assertEqual(obj.do_math(a=4, b=3), 7)

    def test_with_unbound_method_and_op(self):
        """Testing SpyOpMatchInOrder with unbound method and op"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 3,
                    },
                    'op': SpyOpReturn(123),
                },
            ]))

        obj = MathClass()

        self.assertEqual(obj.do_math(a=4, b=3), 123)

    def test_with_expected_calls(self):
        """Testing SpyOpMatchInOrder with all expected calls"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 7,
                    },
                },
                {
                    'kwargs': {
                        'a': 2,
                        'b': 8,
                    },
                    'call_original': False,
                },
                {
                    'kwargs': {
                        'a': 100,
                        'b': 200,
                    },
                    'op': SpyOpReturn(123),
                },
                {
                    'kwargs': {
                        'a': 5,
                        'b': 9,
                    },
                    'call_fake': lambda a, b: a + b + 10,
                },
                {
                    'call_fake': lambda a, b: 1001,
                },
            ]))

        values = [
            obj.do_math(4, 7),
            obj.do_math(a=2, b=8),
            obj.do_math(a=100, b=200),
            obj.do_math(5, b=9),
            obj.do_math(a=1, b=1),
        ]

        self.assertEqual(values, [11, None, 123, 24, 1001])

    def test_with_unexpected_call(self):
        """Testing SpyOpMatchInOrder with unexpected call"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 7,
                    },
                },
            ]))

        expected_message = re.escape(
            "This call to do_math was not passed args=(), "
            "kwargs={'a': 4, 'b': 7}.\n"
            "\n"
            "It was called with:\n"
            "\n"
            "args=()\n"
            "kwargs={'a': 4, 'b': 9}"
        )

        with self.assertRaisesRegex(AssertionError, expected_message):
            obj.do_math(a=4, b=9)

    def test_with_extra_call(self):
        """Testing SpyOpMatchInOrder with extra unexpected call"""
        obj = MathClass()

        self.agency.spy_on(
            obj.do_math,
            op=SpyOpMatchInOrder([
                {
                    'kwargs': {
                        'a': 4,
                        'b': 7,
                    },
                },
            ]))

        self.assertEqual(obj.do_math(a=4, b=7), 11)

        expected_message = re.escape(
            "do_math was called 2 time(s), but only 1 call(s) were expected. "
            "Latest call: <SpyCall(args=(), kwargs={'a': 4, 'b': 9}, "
            "returned=None, raised=None)>"
        )

        with self.assertRaisesRegex(UnexpectedCallError, expected_message):
            obj.do_math(a=4, b=9)


class SpyOpRaiseTests(TestCase):
    """Unit tests for kgb.ops.SpyOpRaise."""

    def test_with_function(self):
        """Testing SpyOpRaise with function"""
        def do_math(a, b):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpRaise(ValueError('foo')))

        with self.assertRaisesRegex(ValueError, 'foo'):
            do_math(5, 3)

    def test_with_classmethod(self):
        """Testing SpyOpRaise with classmethod"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpRaise(ValueError('foo')))

        with self.assertRaisesRegex(ValueError, 'foo'):
            MathClass.class_do_math(5, 3)

    def test_with_unbound_method(self):
        """Testing SpyOpRaise with unbound method"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpRaise(ValueError('foo')))

        obj = MathClass()

        with self.assertRaisesRegex(ValueError, 'foo'):
            obj.do_math(a=4, b=3)


class SpyOpRaiseInOrderTests(TestCase):
    """Unit tests for kgb.ops.SpyOpRaiseInOrder."""

    def test_with_function(self):
        """Testing SpyOpRaiseInOrder with function"""
        def do_math(a, b):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpRaiseInOrder([
                ValueError('foo'),
                KeyError('bar'),
                AttributeError('foobar'),
            ]))

        with self.assertRaisesRegex(ValueError, 'foo'):
            do_math(5, 3)

        with self.assertRaisesRegex(KeyError, 'bar'):
            do_math(5, 3)

        with self.assertRaisesRegex(AttributeError, 'foobar'):
            do_math(5, 3)

        message = re.escape(
            "do_math was called 4 time(s), but only 3 call(s) were expected. "
            "Latest call: <SpyCall(args=(5, 3), kwargs={}, returned=None, "
            "raised=None)>"
        )

        with self.assertRaisesRegex(UnexpectedCallError, message):
            do_math(5, 3)

    def test_with_classmethod(self):
        """Testing SpyOpRaiseInOrder with classmethod"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpRaiseInOrder([
                ValueError('foo'),
                KeyError('bar'),
                AttributeError('foobar'),
            ]))

        with self.assertRaisesRegex(ValueError, 'foo'):
            MathClass.class_do_math(5, 3)

        with self.assertRaisesRegex(KeyError, 'bar'):
            MathClass.class_do_math(5, 3)

        with self.assertRaisesRegex(AttributeError, 'foobar'):
            MathClass.class_do_math(5, 3)

        message = re.escape(
            "class_do_math was called 4 time(s), but only 3 call(s) were "
            "expected. Latest call: <SpyCall(args=(), kwargs={'a': 5, "
            "'b': 3}, returned=None, raised=None)>"
        )

        with self.assertRaisesRegex(UnexpectedCallError, message):
            MathClass.class_do_math(5, 3)

    def test_with_unbound_method(self):
        """Testing SpyOpRaiseInOrder with unbound method"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpRaiseInOrder([
                ValueError('foo'),
                KeyError('bar'),
                AttributeError('foobar'),
            ]))

        obj = MathClass()

        with self.assertRaisesRegex(ValueError, 'foo'):
            obj.do_math(a=4, b=3)

        with self.assertRaisesRegex(KeyError, 'bar'):
            obj.do_math(a=4, b=3)

        with self.assertRaisesRegex(AttributeError, 'foobar'):
            obj.do_math(a=4, b=3)

        message = re.escape(
            "do_math was called 4 time(s), but only 3 call(s) were expected. "
            "Latest call: <SpyCall(args=(), kwargs={'a': 5, 'b': 3}, "
            "returned=None, raised=None)>"
        )

        with self.assertRaisesRegex(UnexpectedCallError, message):
            obj.do_math(5, 3)


class SpyOpReturnTests(TestCase):
    """Unit tests for kgb.ops.SpyOpReturn."""

    def test_with_function(self):
        """Testing SpyOpReturn with function"""
        def do_math(a, b):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpReturn('abc123'))

        self.assertEqual(do_math(5, 3), 'abc123')

    def test_with_classmethod(self):
        """Testing SpyOpReturn with classmethod"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpReturn('abc123'))

        self.assertEqual(MathClass.class_do_math(5, 3), 'abc123')

    def test_with_unbound_method(self):
        """Testing SpyOpReturn with unbound method"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpReturn('abc123'))

        obj = MathClass()

        self.assertEqual(obj.do_math(a=4, b=3), 'abc123')


class SpyOpReturnInOrderTests(TestCase):
    """Unit tests for kgb.ops.SpyOpReturnInOrder."""

    def test_with_function(self):
        """Testing SpyOpReturnInOrder with function"""
        def do_math(a, b):
            return a + b

        self.agency.spy_on(
            do_math,
            op=SpyOpReturnInOrder([
                'abc123',
                'def456',
                'ghi789',
            ]))

        self.assertEqual(do_math(5, 3), 'abc123')
        self.assertEqual(do_math(5, 3), 'def456')
        self.assertEqual(do_math(5, 3), 'ghi789')

        message = re.escape(
            "do_math was called 4 time(s), but only 3 call(s) were expected. "
            "Latest call: <SpyCall(args=(5, 3), kwargs={}, returned=None, "
            "raised=None)>"
        )

        with self.assertRaisesRegex(UnexpectedCallError, message):
            do_math(5, 3)

    def test_with_classmethod(self):
        """Testing SpyOpReturnInOrder with classmethod"""
        self.agency.spy_on(
            MathClass.class_do_math,
            owner=MathClass,
            op=SpyOpReturnInOrder([
                'abc123',
                'def456',
                'ghi789',
            ]))

        self.assertEqual(MathClass.class_do_math(5, 3), 'abc123')
        self.assertEqual(MathClass.class_do_math(5, 3), 'def456')
        self.assertEqual(MathClass.class_do_math(5, 3), 'ghi789')

        message = re.escape(
            "class_do_math was called 4 time(s), but only 3 call(s) were "
            "expected. Latest call: <SpyCall(args=(), kwargs={'a': 5, "
            "'b': 3}, returned=None, raised=None)>"
        )

        with self.assertRaisesRegex(UnexpectedCallError, message):
            MathClass.class_do_math(5, 3)

    def test_with_unbound_method(self):
        """Testing SpyOpReturnInOrder with unbound method"""
        self.agency.spy_on(
            MathClass.do_math,
            owner=MathClass,
            op=SpyOpReturnInOrder([
                'abc123',
                'def456',
                'ghi789',
            ]))

        obj = MathClass()

        self.assertEqual(obj.do_math(a=4, b=3), 'abc123')
        self.assertEqual(obj.do_math(a=4, b=3), 'def456')
        self.assertEqual(obj.do_math(a=4, b=3), 'ghi789')

        message = re.escape(
            "do_math was called 4 time(s), but only 3 call(s) were "
            "expected. Latest call: <SpyCall(args=(), kwargs={'a': 4, "
            "'b': 3}, returned=None, raised=None)>"
        )

        with self.assertRaisesRegex(UnexpectedCallError, message):
            obj.do_math(a=4, b=3)



"""Standalone context managers for working with spies."""

from __future__ import unicode_literals

from contextlib import contextmanager

from kgb.agency import SpyAgency


@contextmanager
def spy_on(*args, **kwargs):
    """Spy on a function.

    By default, the spy will allow the call to go through to the original
    function. This can be disabled by passing ``call_original=False`` when
    initiating the spy. If disabled, the original function will never be
    called.

    This can also be passed a ``call_fake`` parameter pointing to another
    function to call instead of the original. If passed, this will take
    precedence over ``call_original``.

    The spy will only remain throughout the duration of the context.

    See :py:class:`~kgb.spies.FunctionSpy` for more details on arguments.

    Args:
        *args (tuple):
            Positional arguments to pass to
            :py:class:`~kgb.spies.FunctionSpy`.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:class:`~kgb.spies.FunctionSpy`.

    Context:
        kgb.spies.FunctionSpy:
        The newly-created spy.
    """
    agency = SpyAgency()
    spy = agency.spy_on(*args, **kwargs)

    try:
        yield spy
    finally:
        spy.unspy()



from __future__ import unicode_literals

from kgb.pycompat import text_type
from kgb.tests.base import MathClass, TestCase


class SpyCallTests(TestCase):
    """Test cases for kgb.spies.SpyCall."""

    def test_called_with(self):
        """Testing SpyCall.called_with"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(1, b=2)
        obj.do_math_mixed(3, b=4)

        call = obj.do_math_mixed.calls[0]
        self.assertTrue(call.called_with(1, b=2))
        self.assertTrue(call.called_with(a=1, b=2))
        self.assertFalse(call.called_with(3, b=4))
        self.assertFalse(call.called_with(1, 2))

    def test_called_with_and_keyword_args(self):
        """Testing SpyCall.called_with and keyword arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(a=1, b=2)
        obj.do_math_mixed(a=3, b=4)

        call = obj.do_math_mixed.calls[0]
        self.assertTrue(call.called_with(1, b=2))
        self.assertTrue(call.called_with(a=1, b=2))
        self.assertFalse(call.called_with(1, 2))
        self.assertFalse(call.called_with(3, b=4))

    def test_called_with_and_partial_args(self):
        """Testing SpyCall.called_with and partial arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(1, 2)
        obj.do_math_mixed(3, 4)

        call = obj.do_math_mixed.calls[0]
        self.assertTrue(call.called_with(1))
        self.assertFalse(call.called_with(1, 2, 3))
        self.assertFalse(call.called_with(3))

    def test_called_with_and_partial_kwargs(self):
        """Testing SpyCall.called_with and partial keyword arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(a=1, b=2)
        obj.do_math_mixed(a=3, b=4)

        call = obj.do_math_mixed.calls[0]
        self.assertTrue(call.called_with(1))
        self.assertTrue(call.called_with(b=2))
        self.assertTrue(call.called_with(a=1))
        self.assertFalse(call.called_with(a=4))
        self.assertFalse(call.called_with(a=1, b=2, c=3))
        self.assertFalse(call.called_with(a=3, b=2))

    def test_returned(self):
        """Testing SpyCall.returned"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(1, 2)
        obj.do_math_mixed(3, 4)

        call = obj.do_math_mixed.calls[0]
        self.assertTrue(call.returned(3))
        self.assertFalse(call.returned(7))
        self.assertFalse(call.returned(None))

    def test_raised(self):
        """Testing SpyCall.raised"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        with self.assertRaises(TypeError):
            obj.do_math(1, 'a')

        call = obj.do_math.calls[0]
        self.assertTrue(call.raised(TypeError))
        self.assertFalse(call.raised(ValueError))
        self.assertFalse(call.raised(None))

    def test_raised_with_message(self):
        """Testing SpyCall.raised_with_message"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        with self.assertRaises(TypeError):
            obj.do_math(1, 'a')

        call = obj.do_math.calls[0]
        self.assertTrue(call.raised_with_message(
            TypeError,
            "unsupported operand type(s) for +: 'int' and '%s'"
            % text_type.__name__))
        self.assertFalse(call.raised_with_message(
            ValueError,
            "unsupported operand type(s) for +: 'int' and '%s'"
            % text_type.__name__))
        self.assertFalse(call.raised_with_message(TypeError, None))



"""Call tracking and checks for spiess."""

from __future__ import unicode_literals

from kgb.pycompat import iteritems, text_type
from kgb.signature import FunctionSig
from kgb.utils import format_spy_kwargs


class SpyCall(object):
    """Records arguments made to a spied function call.

    SpyCalls are created and stored by a FunctionSpy every time it is
    called. They're accessible through the FunctionSpy's ``calls`` attribute.
    """

    def __init__(self, spy, args, kwargs):
        """Initialize the call.

        Args:
            spy (kgb.spies.FunctionSpy):
                The function spy that the call was made on.

            args (tuple):
                A tuple of positional arguments from the spy. These correspond
                to positional arguments in the function's signature.

            kwargs (dict):
                A dictionary of keyword arguments from the spy. These
                correspond to keyword arguments in the function's signature.
        """
        self.spy = spy
        self.args = args
        self.kwargs = kwargs
        self.return_value = None
        self.exception = None

    def called_with(self, *args, **kwargs):
        """Return whether this call was made with the given arguments.

        Not every argument and keyword argument made in the call must be
        provided to this method. These can be a subset of the positional and
        keyword arguments in the call, but cannot contain any arguments not
        made in the call.

        Args:
            *args (tuple):
                The positional arguments made in the call, or a subset of
                those arguments (starting with the first argument).

            **kwargs (dict):
                The keyword arguments made in the call, or a subset of those
                arguments.

        Returns:
            bool:
            ``True`` if the call's arguments match the provided arguments.
            ``False`` if they do not.
        """
        if len(args) > len(self.args):
            return False

        if self.args[:len(args)] != args:
            return False

        pos_args = self.spy._sig.arg_names

        if self.spy.func_type in (FunctionSig.TYPE_BOUND_METHOD,
                                  FunctionSig.TYPE_UNBOUND_METHOD):
            pos_args = pos_args[1:]

        all_args = dict(zip(pos_args, self.args))
        all_args.update(self.kwargs)

        for key, value in iteritems(kwargs):
            if key not in all_args or all_args[key] != value:
                return False

        return True

    def returned(self, value):
        """Return whether this call returned the given value.

        Args:
            value (object):
                The expected returned value from the call.

        Returns:
            bool:
            ``True`` if this call returned the given value. ``False`` if it
            did not.
        """
        return self.return_value == value

    def raised(self, exception_cls):
        """Return whether this call raised this exception.

        Args:
            exception_cls (type):
                The expected type of exception raised by the call.

        Returns:
            bool:
            ``True`` if this call raised the given exception type.
            ``False`` if it did not.
        """
        return ((self.exception is None and exception_cls is None) or
                type(self.exception) is exception_cls)

    def raised_with_message(self, exception_cls, message):
        """Return whether this call raised this exception and message.

        Args:
            exception_cls (type):
                The expected type of exception raised by the call.

            message (unicode):
                The expected message from the exception.

        Returns:
            bool:
            ``True`` if this call raised the given exception type and message.
            ``False`` if it did not.
        """
        return (self.exception is not None and
                self.raised(exception_cls) and
                text_type(self.exception) == message)

    def __repr__(self):
        return '<SpyCall(args=%r, kwargs=%s, returned=%r, raised=%r)>' % (
            self.args, format_spy_kwargs(self.kwargs), self.return_value,
            self.exception)



from __future__ import unicode_literals

import functools
import inspect
import re
import sys
import traceback
import types
import unittest
import warnings
from contextlib import contextmanager

from kgb.errors import ExistingSpyError, IncompatibleFunctionError
from kgb.pycompat import text_type
from kgb.signature import FunctionSig
from kgb.tests.base import MathClass, TestCase


has_getargspec = hasattr(inspect, 'getargspec')
has_getfullargspec = hasattr(inspect, 'getfullargspec')


def require_getargspec(func):
    """Require getargspec for a unit test.

    If not available, the test will be skipped.

    Args:
        func (callable):
            The unit test function to decorate.
    """
    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        if has_getargspec:
            with warnings.catch_warnings(record=True):
                return func(*args, **kwargs)
        else:
            raise unittest.SkipTest(
                'inspect.getargspec is not available on Python %s.%s.%s'
                % sys.version_info[:3])

    return _wrap


def do_math(a=1, b=2, *args, **kwargs):
    return a - b


def do_math_pos(a, b):
    return a - b


def do_math_mixed(a, b=2, *args, **kwargs):
    return a - b


def fake_do_math(self, a, b, *args, **kwargs):
    assert isinstance(self, MathClass)

    return a - b


def fake_class_do_math(cls, a, b, *args, **kwargs):
    assert issubclass(cls, MathClass)

    return a - b


def something_awesome():
    return 'Tada!'


def fake_something_awesome():
    return r'\o/'


@contextmanager
def do_context(a=1, b=2):
    yield a + b


class AdderObject(object):
    def func(self):
        assert isinstance(self, AdderObject)

        return [self.add_one(i) for i in (1, 2, 3)]

    def add_one(self, i):
        assert isinstance(self, AdderObject)

        return i + 1

    @classmethod
    def class_func(cls):
        assert cls is AdderObject

        return [cls.class_add_one(i) for i in (1, 2, 3)]

    @classmethod
    def class_add_one(cls, i):
        assert issubclass(cls, AdderObject)

        return i + 1


class AdderSubclass(AdderObject):
    pass


class slippery_func(object):
    count = 0

    def __init__(self):
        pass

    def __call__(self, method):
        self.method = method
        return self

    def __get__(self, obj, objtype=None):
        @functools.wraps(self.method)
        def _wrapper(*args, **kwargs):
            return self.method(obj, _wrapper.counter)

        _wrapper.counter = self.count
        type(self).count += 1

        return _wrapper


class SlipperyFuncObject(object):
    @slippery_func()
    def my_func(self, value):
        return value


class FunctionSpyTests(TestCase):
    """Test cases for kgb.spies.FunctionSpy."""

    def test_construction_with_call_precedence(self):
        """Testing FunctionSpy construction with call option precedence"""
        spy = self.agency.spy_on(something_awesome,
                                 call_fake=fake_something_awesome,
                                 call_original=True)
        self.assertEqual(spy.func, fake_something_awesome)

    def test_construction_with_call_fake(self):
        """Testing FunctionSpy construction with call_fake"""
        spy = self.agency.spy_on(something_awesome,
                                 call_fake=fake_something_awesome)

        self.assertTrue(hasattr(something_awesome, 'spy'))
        self.assertEqual(something_awesome.spy, spy)
        self.assertEqual(
            getattr(spy.func, FunctionSig.FUNC_NAME_ATTR),
            getattr(fake_something_awesome, FunctionSig.FUNC_NAME_ATTR))
        self.assertEqual(spy.orig_func, something_awesome)
        self.assertEqual(spy.func_name, 'something_awesome')
        self.assertEqual(spy.func_type, spy.TYPE_FUNCTION)
        self.assertIsInstance(something_awesome, types.FunctionType)

    def test_construction_with_call_fake_and_bound_method(self):
        """Testing FunctionSpy construction with call_fake and bound method"""
        obj = MathClass()
        orig_method = obj.do_math
        spy = self.agency.spy_on(obj.do_math, call_fake=fake_do_math)

        self.assertTrue(hasattr(obj.do_math, 'spy'))
        self.assertIs(obj.do_math.spy, spy)
        self.assertIsNot(obj.do_math, orig_method)
        self.assertFalse(hasattr(MathClass.do_math, 'spy'))

        self.assertEqual(spy.func, fake_do_math)
        self.assertEqual(spy.func_name, 'do_math')
        self.assertEqual(spy.func_type, spy.TYPE_BOUND_METHOD)
        self.assertIsInstance(obj.do_math, types.MethodType)

    def test_construction_with_call_fake_and_unbound_method(self):
        """Testing FunctionSpy construction with call_fake and unbound method
        """
        orig_method = MathClass.do_math
        spy = self.agency.spy_on(MathClass.do_math, call_fake=fake_do_math)

        self.assertTrue(hasattr(MathClass.do_math, 'spy'))
        self.assertIs(MathClass.do_math.spy, spy)
        self.assertEqual(spy.func, fake_do_math)
        self.assertEqual(spy.func_name, 'do_math')
        self.assertEqual(spy.func_type, spy.TYPE_UNBOUND_METHOD)
        self.assertEqual(spy.owner, MathClass)

        if isinstance(orig_method, types.FunctionType):
            # Python 3
            self.assertIs(MathClass.do_math, orig_method)
        elif isinstance(orig_method, types.MethodType):
            # Python 2
            self.assertIsNot(MathClass.do_math, orig_method)
        else:
            self.fail('Method has an unexpected type %r' % type(orig_method))

        obj = MathClass()
        self.assertTrue(hasattr(obj.do_math, 'spy'))
        self.assertIs(obj.do_math.spy, MathClass.do_math.spy)
        self.assertIsInstance(obj.do_math, types.MethodType)

    def test_construction_with_call_fake_and_classmethod(self):
        """Testing FunctionSpy construction with call_fake and classmethod"""
        def fake_class_do_math(cls, *args, **kwargs):
            return 42

        orig_method = MathClass.class_do_math
        spy = self.agency.spy_on(MathClass.class_do_math,
                                 call_fake=fake_class_do_math)

        self.assertTrue(hasattr(MathClass.class_do_math, 'spy'))
        self.assertIs(MathClass.class_do_math.spy, spy)
        self.assertIs(MathClass.class_do_math, orig_method)

        self.assertEqual(spy.func, fake_class_do_math)
        self.assertEqual(spy.orig_func, self.orig_class_do_math)
        self.assertEqual(spy.func_name, 'class_do_math')
        self.assertIsInstance(MathClass.class_do_math, types.MethodType)

    def test_construction_with_call_original_false(self):
        """Testing FunctionSpy construction with call_original=False"""
        obj = MathClass()
        spy = self.agency.spy_on(obj.do_math, call_original=False)

        self.assertTrue(hasattr(obj.do_math, 'spy'))
        self.assertIs(obj.do_math.spy, spy)

        self.assertIsNone(spy.func)
        self.assertEqual(spy.func_name, 'do_math')
        self.assertIsInstance(obj.do_math, types.MethodType)

    def test_construction_with_call_original_true(self):
        """Testing FunctionSpy construction with call_original=True"""
        spy = self.agency.spy_on(something_awesome, call_original=True)

        self.assertTrue(hasattr(something_awesome, 'spy'))
        self.assertEqual(something_awesome.spy, spy)
        self.assertEqual(
            getattr(spy.func, FunctionSig.FUNC_NAME_ATTR),
            getattr(something_awesome, FunctionSig.FUNC_NAME_ATTR))
        self.assertEqual(spy.orig_func, something_awesome)
        self.assertEqual(spy.func_name, 'something_awesome')
        self.assertIsInstance(something_awesome, types.FunctionType)

    def test_construction_with_call_original_true_and_bound_method(self):
        """Testing FunctionSpy construction with call_original=True and bound
        method
        """
        obj = MathClass()
        orig_do_math = obj.do_math

        spy = self.agency.spy_on(obj.do_math, call_original=True)

        self.assertTrue(hasattr(obj.do_math, 'spy'))
        self.assertIs(obj.do_math.spy, spy)
        self.assertFalse(hasattr(MathClass.do_math, 'spy'))

        self.assertEqual(spy.orig_func, orig_do_math)
        self.assertEqual(spy.func_name, 'do_math')
        self.assertEqual(spy.func_type, spy.TYPE_BOUND_METHOD)
        self.assertIsInstance(obj.do_math, types.MethodType)

    def test_construction_with_call_original_and_classmethod(self):
        """Testing FunctionSpy construction with call_original and classmethod
        """
        spy = self.agency.spy_on(MathClass.class_do_math, call_original=True)

        self.assertTrue(hasattr(MathClass.class_do_math, 'spy'))
        self.assertIs(MathClass.class_do_math.spy, spy)

        self.assertEqual(spy.orig_func, self.orig_class_do_math)
        self.assertEqual(spy.func_name, 'class_do_math')
        self.assertEqual(spy.func_type, spy.TYPE_BOUND_METHOD)
        self.assertIsInstance(MathClass.class_do_math, types.MethodType)

    def test_construction_with_function_and_owner(self):
        """Testing FunctionSpy constructions with function and owner passed"""
        with self.assertRaises(ValueError) as cm:
            self.agency.spy_on(do_math, owner=AdderObject)

        self.assertEqual(text_type(cm.exception),
                         'This function has no owner, but an owner was '
                         'passed to spy_on().')

    def test_init_with_slippery_bound_method(self):
        """Testing FunctionSpy construction with new slippery function
        generated dynamically from accessing a spied-on bound method from
        instance
        """
        obj = SlipperyFuncObject()

        # Verify that we're producing a different function on each access.
        func1 = obj.my_func
        func2 = obj.my_func
        self.assertIsInstance(func1, types.FunctionType)
        self.assertIsInstance(func2, types.FunctionType)
        self.assertIsNot(func1, func2)

        spy = self.agency.spy_on(obj.my_func,
                                 owner=obj,
                                 call_original=True)

        self.assertTrue(hasattr(obj.my_func, 'spy'))
        self.assertIs(obj.my_func.spy, spy)

        self.assertEqual(spy.func_name, 'my_func')
        self.assertEqual(spy.func_type, spy.TYPE_BOUND_METHOD)
        self.assertIsInstance(obj.my_func, types.MethodType)

        slippery_func.count = 0
        func1 = obj.my_func
        func2 = obj.my_func
        self.assertIs(func1, func2)

        # Since the functions are wrapped, we'll be getting a new value each
        # time we invoke them (and not when accessing the attributes as we
        # did above).
        self.assertEqual(func1(), 0)
        self.assertEqual(func1(), 1)
        self.assertEqual(func2(), 2)
        self.assertEqual(len(obj.my_func.calls), 3)

    def test_init_with_slippery_unbound_method(self):
        """Testing FunctionSpy construction with new slippery function
        generated dynamically from accessing a spied-on unbound method from
        instance
        """
        with self.assertRaises(ValueError) as cm:
            self.agency.spy_on(SlipperyFuncObject.my_func,
                               owner=SlipperyFuncObject)

        self.assertEqual(
            text_type(cm.exception),
            'Unable to spy on unbound slippery methods (those that return '
            'a new function on each attribute access). Please spy on an '
            'instance instead.')

    def test_construction_with_classmethod_on_parent(self):
        """Testing FunctionSpy construction with classmethod from parent of
        class
        """
        class MyParent(object):
            @classmethod
            def foo(self):
                pass

        class MyObject(MyParent):
            pass

        obj = MyObject()
        orig_method = obj.foo

        spy = self.agency.spy_on(MyObject.foo)

        self.assertTrue(hasattr(MyObject.foo, 'spy'))
        self.assertFalse(hasattr(MyParent.foo, 'spy'))
        self.assertIs(MyObject.foo.spy, spy)
        self.assertEqual(spy.func_name, 'foo')
        self.assertEqual(spy.func_type, spy.TYPE_BOUND_METHOD)
        self.assertEqual(spy.owner, MyObject)

        if isinstance(orig_method, types.FunctionType):
            # Python 3
            self.assertIs(MyObject.foo, orig_method)
        elif isinstance(orig_method, types.MethodType):
            # Python 2
            self.assertIsNot(MyObject.foo, orig_method)
        else:
            self.fail('Method has an unexpected type %r' % type(orig_method))

        obj2 = MyObject()
        self.assertTrue(hasattr(obj2.foo, 'spy'))
        self.assertIs(obj2.foo.spy, MyObject.foo.spy)
        self.assertIsInstance(obj2.foo, types.MethodType)

        obj3 = MyParent()
        self.assertFalse(hasattr(obj3.foo, 'spy'))

    def test_construction_with_unbound_method_on_parent(self):
        """Testing FunctionSpy construction with unbound method from parent of
        class
        """
        obj = AdderSubclass()
        orig_method = obj.func

        spy = self.agency.spy_on(AdderSubclass.func, owner=AdderSubclass)

        self.assertTrue(hasattr(AdderSubclass.func, 'spy'))
        self.assertFalse(hasattr(AdderObject.func, 'spy'))
        self.assertIs(AdderSubclass.func.spy, spy)
        self.assertEqual(spy.func_name, 'func')
        self.assertEqual(spy.func_type, spy.TYPE_UNBOUND_METHOD)
        self.assertEqual(spy.owner, AdderSubclass)

        if isinstance(orig_method, types.FunctionType):
            # Python 3
            self.assertIs(AdderSubclass.func, orig_method)
        elif isinstance(orig_method, types.MethodType):
            # Python 2
            self.assertIsNot(AdderSubclass.func, orig_method)
        else:
            self.fail('Method has an unexpected type %r' % type(orig_method))

        obj2 = AdderSubclass()
        self.assertTrue(hasattr(obj2.func, 'spy'))
        self.assertIs(obj2.func.spy, AdderSubclass.func.spy)
        self.assertIsInstance(obj2.func, types.MethodType)

        obj3 = AdderObject()
        self.assertFalse(hasattr(obj3.func, 'spy'))

    def test_construction_with_falsy_im_self(self):
        """Testing FunctionSpy construction with a falsy function.im_self"""
        class MyObject(dict):
            def foo(self):
                pass

        my_object = MyObject()
        orig_foo = my_object.foo

        # Ensure it's falsy.
        self.assertFalse(my_object)

        spy = self.agency.spy_on(my_object.foo)

        self.assertEqual(spy.orig_func, orig_foo)
        self.assertNotEqual(MyObject.foo, spy)
        self.assertEqual(spy.func_name, 'foo')
        self.assertEqual(spy.func_type, spy.TYPE_BOUND_METHOD)
        self.assertIsInstance(my_object.foo, types.MethodType)

    def test_construction_with_existing_spy(self):
        """Testing FunctionSpy constructions with function already spied on"""
        def setup_spy():
            self.agency.spy_on(do_math)

        setup_spy()

        with self.assertRaises(ExistingSpyError) as cm:
            self.agency.spy_on(do_math)

        self.assertIn(', in setup_spy', text_type(cm.exception))

    def test_construction_with_bound_method_and_custom_setattr(self):
        """Testing FunctionSpy constructions with a bound method on a class
        containing a custom __setattr__
        """
        class MyObject(object):
            def __setattr__(self, key, value):
                assert False

            def foo(self):
                pass

        obj = MyObject()
        orig_foo = obj.foo

        spy = self.agency.spy_on(obj.foo)
        self.assertEqual(spy.orig_func, orig_foo)
        self.assertNotEqual(MyObject.foo, spy)
        self.assertEqual(spy.func_name, 'foo')
        self.assertEqual(spy.func_type, spy.TYPE_BOUND_METHOD)
        self.assertIsInstance(obj.foo, types.MethodType)
        self.assertTrue(hasattr(obj.foo, 'spy'))
        self.assertTrue(hasattr(obj.foo, 'called_with'))

        obj2 = MyObject()
        self.assertFalse(hasattr(obj2.foo, 'spy'))

    def test_construction_with_bound_method_and_bad_owner(self):
        """Testing FunctionSpy constructions with a bound method and an
        explicit owner not matching the class
        """
        class MyObject(object):
            def foo(self):
                pass

        class BadObject(object):
            def foo(self):
                pass

        obj = MyObject()

        with self.assertRaises(ValueError) as cm:
            self.agency.spy_on(obj.foo, owner=BadObject)

        self.assertEqual(text_type(cm.exception),
                         'The owner passed does not match the actual owner '
                         'of the bound method.')

    def test_construction_with_owner_without_method(self):
        """Testing FunctionSpy constructions with an owner passed that does
        not provide the spied method
        """
        class MyObject(object):
            def foo(self):
                pass

        obj = MyObject()

        with self.assertRaises(ValueError) as cm:
            self.agency.spy_on(obj.foo, owner=AdderObject)

        self.assertEqual(text_type(cm.exception),
                         'The owner passed does not contain the spied method.')

    def test_construction_with_non_function(self):
        """Testing FunctionSpy constructions with non-function"""
        with self.assertRaises(ValueError) as cm:
            self.agency.spy_on(42)

        self.assertEqual(text_type(cm.exception),
                         '42 cannot be spied on. It does not appear to be a '
                         'valid function or method.')

    def test_construction_with_call_fake_non_function(self):
        """Testing FunctionSpy constructions with call_fake as non-function"""
        with self.assertRaises(ValueError) as cm:
            self.agency.spy_on(do_math, call_fake=True)

        self.assertEqual(text_type(cm.exception),
                         'True cannot be used for call_fake. It does not '
                         'appear to be a valid function or method.')

    def test_construction_with_call_fake_compatibility(self):
        """Testing FunctionSpy constructions with call_fake with signature
        compatibility
        """
        def source1(a, b):
            pass

        def source2(a, b, *args):
            pass

        def source3(c=1, d=2):
            pass

        def source4(c=1, d=2, **kwargs):
            pass

        with self.assertRaises(IncompatibleFunctionError):
            self.agency.spy_on(
                source1,
                call_fake=lambda a, b, c: None)

        with self.assertRaises(IncompatibleFunctionError):
            self.agency.spy_on(
                source1,
                call_fake=lambda a: None)

        with self.assertRaises(IncompatibleFunctionError):
            self.agency.spy_on(
                source1,
                call_fake=lambda **kwargs: None)

        with self.assertRaises(IncompatibleFunctionError):
            self.agency.spy_on(
                source2,
                call_fake=lambda a, b: None)

        with self.assertRaises(IncompatibleFunctionError):
            self.agency.spy_on(
                source3,
                call_fake=lambda c=1: None)

        with self.assertRaises(IncompatibleFunctionError):
            self.agency.spy_on(
                source4,
                call_fake=lambda c=1, d=2, e=3: None)

        with self.assertRaises(IncompatibleFunctionError):
            self.agency.spy_on(
                source4,
                call_fake=lambda c=1, d=2: None)

        self.agency.spy_on(source1, call_fake=lambda a, b: None)
        source1.unspy()

        self.agency.spy_on(source1, call_fake=lambda *args: None)
        source1.unspy()

        self.agency.spy_on(source4, call_fake=lambda c=1, d=2, **kwargs: None)
        source4.unspy()

        self.agency.spy_on(source4, call_fake=lambda c=1, **kwargs: None)
        source4.unspy()

        self.agency.spy_on(source4, call_fake=lambda c, d=None, **kwargs: None)
        source4.unspy()

        self.agency.spy_on(source4, call_fake=lambda c, e, **kwargs: None)
        source4.unspy()

        self.agency.spy_on(source4, call_fake=lambda **kwargs: None)
        source4.unspy()

    def test_construction_with_old_style_class(self):
        """Testing FunctionSpy with old-style class"""
        class MyClass:
            def test_func(self):
                return 100

        obj = MyClass()

        self.agency.spy_on(obj.test_func, call_fake=lambda obj: 200)
        self.assertEqual(obj.test_func(), 200)

    def test_call_with_fake(self):
        """Testing FunctionSpy calls with call_fake"""
        self.agency.spy_on(something_awesome,
                           call_fake=fake_something_awesome)
        result = something_awesome()

        self.assertEqual(result, r'\o/')
        self.assertEqual(len(something_awesome.spy.calls), 1)
        self.assertEqual(len(something_awesome.spy.calls[0].args), 0)
        self.assertEqual(len(something_awesome.spy.calls[0].kwargs), 0)

    def test_call_with_fake_and_bound_method(self):
        """Testing FunctionSpy calls with call_fake and bound method"""
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_fake=fake_do_math)
        result = obj.do_math()

        self.assertEqual(result, -1)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertTrue(obj.do_math.last_called_with(
            a=1,
            b=2))

    def test_call_with_fake_and_unbound_method(self):
        """Testing FunctionSpy calls with call_fake and unbound method"""
        self.agency.spy_on(MathClass.do_math, call_fake=fake_do_math)

        obj = MathClass()
        result = obj.do_math()

        self.assertEqual(result, -1)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertTrue(obj.do_math.last_called_with(
            a=1,
            b=2))

    def test_call_with_fake_and_classmethod(self):
        """Testing FunctionSpy calls with call_fake and classmethod"""
        self.agency.spy_on(MathClass.class_do_math,
                           call_fake=fake_class_do_math)
        result = MathClass.class_do_math()

        self.assertEqual(result, -3)
        self.assertEqual(len(MathClass.class_do_math.calls), 1)
        self.assertTrue(MathClass.class_do_math.last_called_with(
            a=2,
            b=5))

    def test_call_with_fake_and_contextmanager(self):
        """Testing FunctionSpy calls with call_fake and context manager"""
        @contextmanager
        def fake_do_context(a=1, b=2):
            yield a * b

        self.agency.spy_on(do_context,
                           call_fake=fake_do_context)

        with do_context(a=5) as ctx:
            result = ctx

        self.assertEqual(result, 10)
        self.assertEqual(len(do_context.calls), 1)
        self.assertEqual(do_context.calls[0].args, ())
        self.assertEqual(do_context.calls[0].kwargs, {'a': 5})

    def test_call_with_fake_and_contextmanager_func_raises_exception(self):
        """Testing FunctionSpy calls with call_fake and context manager and
        function raises exception
        """
        e = Exception('oh no')

        @contextmanager
        def fake_do_context(*args, **kwargs):
            raise e

        self.agency.spy_on(do_context,
                           call_fake=fake_do_context)

        with self.assertRaisesRegex(Exception, 'oh no'):
            with do_context(a=5):
                pass

        self.assertEqual(len(do_context.calls), 1)
        self.assertEqual(do_context.calls[0].args, ())
        self.assertEqual(do_context.calls[0].kwargs, {'a': 5})
        self.assertEqual(do_context.calls[0].exception, e)

    def test_call_with_fake_and_contextmanager_body_raises_exception(self):
        """Testing FunctionSpy calls with call_fake and context manager and
        context body raises exception
        """
        e = Exception('oh no')

        @contextmanager
        def fake_do_context(a=1, b=2):
            yield a * b

        self.agency.spy_on(do_context,
                           call_fake=fake_do_context)

        with self.assertRaisesRegex(Exception, 'oh no'):
            with do_context(a=5):
                raise e

        self.assertEqual(len(do_context.calls), 1)
        self.assertEqual(do_context.calls[0].args, ())
        self.assertEqual(do_context.calls[0].kwargs, {'a': 5})
        self.assertIsNone(do_context.calls[0].exception)

    def test_call_with_exception(self):
        e = ValueError('oh no')

        def orig_func(arg1=None, arg2=None):
            # Create enough of a difference in code positions between this
            # and the forwarding functions, to ensure the exception's
            # position count is higher than that of the forwarding function.
            #
            # This is important for sanity checks on Python 3.11.
            try:
                if 1:
                    if 2:
                        try:
                            a = 1
                            b = a
                            a = 2
                        except Exception:
                            raise
                    else:
                        c = [1, 2, 3, 4, 5]
                        a = c
            except Exception:
                raise

            for i in range(10):
                try:
                    d = [1, 2, 3, 4, 5]
                    a = d
                    b = a
                    d = b
                except Exception:
                    raise

            # We should be good. We'll verify counts later.
            raise e

        # Verify the above.
        orig_func_code = orig_func.__code__
        supports_co_positions = hasattr(orig_func_code, 'co_positions')

        if supports_co_positions:
            orig_positions_count = len(list(orig_func_code.co_positions()))
        else:
            orig_positions_count = None

        # Now spy.
        self.agency.spy_on(orig_func)

        if supports_co_positions:
            spy_positions_count = len(list(orig_func.__code__.co_positions()))

            # Make sure we had enough padding up above.
            self.assertGreater(orig_positions_count, spy_positions_count)

        # Now test.
        try:
            orig_func()
        except Exception as ex:
            # This should fail if we've built the CodeType wrong and have a
            # resulting offset issue. The act of pretty-printing the exception
            # triggers the noticeable co_positions() issue.
            traceback.print_exception(*sys.exc_info())

        self.assertEqual(len(orig_func.calls), 1)
        self.assertIs(orig_func.calls[0].exception, e)

    def test_call_with_fake_and_args(self):
        """Testing FunctionSpy calls with call_fake and arguments"""
        obj = MathClass()

        self.agency.spy_on(obj.do_math_pos, call_fake=fake_do_math)
        result = obj.do_math_pos(10, 20)

        self.assertEqual(result, -10)
        self.assertEqual(len(obj.do_math_pos.calls), 1)
        self.assertEqual(obj.do_math_pos.calls[0].args, (10, 20))
        self.assertEqual(obj.do_math_pos.calls[0].kwargs, {})

    def test_call_with_fake_and_args_for_kwargs(self):
        """Testing FunctionSpy calls with call_fake and positional arguments
        in place of keyword arguments
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_fake=fake_do_math)
        result = obj.do_math(10, 20)

        self.assertEqual(result, -10)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertEqual(obj.do_math.calls[0].args, ())
        self.assertEqual(obj.do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20,
        })

    def test_call_with_fake_and_kwargs(self):
        """Testing FunctionSpy calls with call_fake and keyword arguments"""
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_fake=fake_do_math)
        result = obj.do_math(a=10, b=20)

        self.assertEqual(result, -10)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertEqual(len(obj.do_math.calls[0].args), 0)
        self.assertEqual(obj.do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20,
        })

    def test_call_with_original_false(self):
        """Testing FunctionSpy calls with call_original=False"""
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_original=False)
        result = obj.do_math()

        self.assertIsNone(result)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertTrue(obj.do_math.last_called_with(a=1, b=2))

    def test_call_with_all_original_false_and_args(self):
        """Testing FunctionSpy calls with call_original=False and positional
        arguments
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math_pos, call_original=False)
        result = obj.do_math_pos(10, 20)

        self.assertIsNone(result)
        self.assertEqual(len(obj.do_math_pos.calls), 1)
        self.assertEqual(obj.do_math_pos.calls[0].args, (10, 20))
        self.assertEqual(obj.do_math_pos.calls[0].kwargs, {})

    def test_call_with_all_original_false_and_args_for_kwargs(self):
        """Testing FunctionSpy calls with call_original=False and positional
        arguments in place of keyword arguments
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_original=False)
        result = obj.do_math(10, 20)

        self.assertEqual(result, None)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertEqual(obj.do_math.calls[0].args, ())
        self.assertEqual(obj.do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20,
        })

    def test_call_with_original_false_and_kwargs(self):
        """Testing FunctionSpy calls with call_original=False and keyword
        arguments
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_original=False)
        result = obj.do_math(a=10, b=20)

        self.assertEqual(result, None)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertEqual(len(obj.do_math.calls[0].args), 0)
        self.assertEqual(obj.do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20
        })

    def test_call_with_original_true_and_function(self):
        """Testing FunctionSpy calls with call_original=True and function"""
        self.agency.spy_on(something_awesome, call_original=True)
        result = something_awesome()

        self.assertEqual(result, 'Tada!')
        self.assertTrue(hasattr(something_awesome, 'spy'))
        self.assertEqual(len(something_awesome.spy.calls), 1)
        self.assertEqual(len(something_awesome.spy.calls[0].args), 0)
        self.assertEqual(len(something_awesome.spy.calls[0].kwargs), 0)

    def test_call_with_original_true_and_function_args(self):
        """Testing FunctionSpy calls with call_original=True and function
        with all positional arguments
        """
        self.agency.spy_on(do_math_pos, call_original=True)
        result = do_math_pos(10, 20)

        self.assertEqual(result, -10)
        self.assertEqual(len(do_math_pos.spy.calls), 1)
        self.assertEqual(do_math_pos.spy.calls[0].args, (10, 20))
        self.assertEqual(len(do_math_pos.spy.calls[0].kwargs), 0)

    def test_call_with_original_true_and_function_args_for_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and function
        with all positional arguments in place of keyword arguments
        """
        self.agency.spy_on(do_math, call_original=True)
        result = do_math(10, 20)

        self.assertEqual(result, -10)
        self.assertEqual(len(do_math.spy.calls), 1)
        self.assertEqual(do_math.spy.calls[0].args, ())
        self.assertEqual(do_math.spy.calls[0].kwargs, {
            'a': 10,
            'b': 20,
        })

    def test_call_with_original_true_and_function_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and function
        with all keyword arguments
        """
        self.agency.spy_on(do_math, call_original=True)
        result = do_math(b=10, a=20)

        self.assertEqual(result, 10)
        self.assertEqual(len(do_math.spy.calls), 1)
        self.assertEqual(len(do_math.spy.calls[0].args), 0)
        self.assertEqual(do_math.spy.calls[0].kwargs, {
            'a': 20,
            'b': 10
        })

    def test_call_with_original_true_and_function_mixed(self):
        """Testing FunctionSpy calls with call_original=True and function
        with all mixed argument types
        """
        self.agency.spy_on(do_math, call_original=True)
        result = do_math(10, b=20, unused=True)

        self.assertEqual(result, -10)
        self.assertEqual(len(do_math.spy.calls), 1)
        self.assertEqual(do_math.spy.calls[0].args, ())
        self.assertEqual(do_math.spy.calls[0].kwargs, {
            'a': 10,
            'b': 20,
            'unused': True,
        })

        self.agency.spy_on(do_math_pos, call_original=True)
        result = do_math_pos(10, b=20)

        self.assertEqual(result, -10)
        self.assertEqual(len(do_math_pos.spy.calls), 1)
        self.assertEqual(do_math_pos.spy.calls[0].args, (10, 20))
        self.assertEqual(do_math_pos.spy.calls[0].kwargs, {})

        self.agency.spy_on(do_math_mixed, call_original=True)
        result = do_math_mixed(10, b=20, unused=True)

        self.assertEqual(result, -10)
        self.assertEqual(len(do_math_mixed.spy.calls), 1)
        self.assertEqual(do_math_mixed.spy.calls[0].args, (10,))
        self.assertEqual(do_math_mixed.spy.calls[0].kwargs, {
            'b': 20,
            'unused': True,
        })

    def test_call_with_original_true_and_bound_method(self):
        """Testing FunctionSpy calls with call_original=True and bound method
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_original=True)
        result = obj.do_math()

        self.assertEqual(result, 3)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertTrue(obj.do_math.last_called_with(a=1, b=2))

    def test_call_with_original_true_and_bound_method_args(self):
        """Testing FunctionSpy calls with call_original=True and bound method
        with all positional arguments
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math_pos, call_original=True)
        result = obj.do_math_pos(10, 20)

        self.assertEqual(result, 30)
        self.assertEqual(len(obj.do_math_pos.calls), 1)
        self.assertEqual(obj.do_math_pos.calls[0].args, (10, 20))
        self.assertEqual(len(obj.do_math_pos.calls[0].kwargs), 0)

    def test_call_with_original_true_and_bound_method_args_for_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and bound method
        with all positional arguments in place of keyword arguments
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_original=True)
        result = obj.do_math(10, 20)

        self.assertEqual(result, 30)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertEqual(obj.do_math.calls[0].args, ())
        self.assertEqual(obj.do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20,
        })

    def test_call_with_original_true_and_bound_method_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and bound method
        with all keyword arguments
        """
        obj = MathClass()

        self.agency.spy_on(obj.do_math, call_original=True)
        result = obj.do_math(a=10, b=20)

        self.assertEqual(result, 30)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertEqual(len(obj.do_math.calls[0].args), 0)
        self.assertEqual(obj.do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20
        })

    def test_call_with_original_true_and_unbound_method(self):
        """Testing FunctionSpy calls with call_original=True and unbound method
        """
        self.agency.spy_on(MathClass.do_math, call_original=True)

        obj = MathClass()
        result = obj.do_math()

        self.assertEqual(result, 3)
        self.assertEqual(len(MathClass.do_math.calls), 1)
        self.assertTrue(MathClass.do_math.last_called_with(a=1, b=2))

    def test_call_with_original_true_and_unbound_method_args(self):
        """Testing FunctionSpy calls with call_original=True and unbound
        method with all positional arguments
        """
        self.agency.spy_on(MathClass.do_math_pos, call_original=True)

        obj = MathClass()
        result = obj.do_math_pos(10, 20)

        self.assertEqual(result, 30)
        self.assertEqual(len(MathClass.do_math_pos.calls), 1)
        self.assertTrue(MathClass.do_math_pos.last_called_with(10, 20))

    def test_call_with_original_true_and_unbound_method_args_for_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and unbound
        method with all positional arguments in place of keyword arguments
        """
        self.agency.spy_on(MathClass.do_math, call_original=True)

        obj = MathClass()
        result = obj.do_math(10, 20)

        self.assertEqual(result, 30)
        self.assertEqual(len(MathClass.do_math.calls), 1)
        self.assertTrue(MathClass.do_math.last_called_with(a=10, b=20))

    def test_call_with_original_true_and_unbound_method_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and unbound method
        with all keyword arguments
        """
        self.agency.spy_on(MathClass.do_math, call_original=True)

        obj = MathClass()
        result = obj.do_math(a=10, b=20)

        self.assertEqual(result, 30)
        self.assertEqual(len(MathClass.do_math.calls), 1)
        self.assertEqual(len(MathClass.do_math.calls[0].args), 0)
        self.assertTrue(MathClass.do_math.last_called_with(a=10, b=20))

    def test_call_with_original_true_and_classmethod(self):
        """Testing FunctionSpy calls with call_original=True and classmethod"""
        self.agency.spy_on(MathClass.class_do_math, call_original=True)
        result = MathClass.class_do_math()

        self.assertEqual(result, 10)
        self.assertEqual(len(MathClass.class_do_math.calls), 1)
        self.assertTrue(MathClass.class_do_math.last_called_with(a=2, b=5))

    def test_call_with_original_true_and_classmethod_args(self):
        """Testing FunctionSpy calls with call_original=True and classmethod
        with all positional arguments
        """
        self.agency.spy_on(MathClass.class_do_math_pos, call_original=True)
        result = MathClass.class_do_math_pos(10, 20)

        self.assertEqual(result, 200)
        self.assertEqual(len(MathClass.class_do_math_pos.calls), 1)
        self.assertEqual(MathClass.class_do_math_pos.calls[0].args, (10, 20))
        self.assertEqual(len(MathClass.class_do_math_pos.calls[0].kwargs), 0)

    def test_call_with_original_true_and_classmethod_args_for_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and classmethod
        with all positional arguments in place of keyword arguments
        """
        self.agency.spy_on(MathClass.class_do_math, call_original=True)
        result = MathClass.class_do_math(10, 20)

        self.assertEqual(result, 200)
        self.assertEqual(len(MathClass.class_do_math.calls), 1)
        self.assertEqual(MathClass.class_do_math.calls[0].args, ())
        self.assertEqual(MathClass.class_do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20,
        })

    def test_call_with_original_true_and_classmethod_kwargs(self):
        """Testing FunctionSpy calls with call_original=True and classmethod
        with all keyword arguments
        """
        self.agency.spy_on(MathClass.class_do_math, call_original=True)
        result = MathClass.class_do_math(a=10, b=20)

        self.assertEqual(result, 200)
        self.assertEqual(len(MathClass.class_do_math.calls), 1)
        self.assertEqual(len(MathClass.class_do_math.calls[0].args), 0)
        self.assertEqual(MathClass.class_do_math.calls[0].kwargs, {
            'a': 10,
            'b': 20
        })

    def test_call_with_inline_function_using_closure_vars(self):
        """Testing FunctionSpy calls for inline function using a closure's
        variables
        """
        d = {}

        def func():
            d['called'] = True

        self.agency.spy_on(func)

        func()
        self.assertTrue(func.called)
        self.assertEqual(d, {'called': True})

    def test_call_with_inline_function_using_closure_vars_and_args(self):
        """Testing FunctionSpy calls for inline function using a closure's
        variables and function with arguments
        """
        d = {}

        def func(a, b, c):
            d['call_result'] = a + b + c

        self.agency.spy_on(func)

        func(1, 2, c=3)
        self.assertTrue(func.called)
        self.assertEqual(d, {'call_result': 6})

    def test_call_with_function_providing_closure_vars(self):
        """Testing FunctionSpy calls for function providing variables for an
        inline function
        """
        def func():
            d = {}

            def inline_func():
                d['called'] = True

            inline_func()

            return d

        self.agency.spy_on(func)

        d = func()
        self.assertTrue(func.called)
        self.assertEqual(d, {'called': True})

    def test_call_with_function_providing_closure_access_args(self):
        """Testing FunctionSpy calls with an inline function accessing
        parent's arguments
        """
        # NOTE: This originally caused a crash on Python 3.13 beta 2.
        #       See: https://github.com/beanbaginc/kgb/issues/11
        def func(arg):
            def inline_func():
                print(arg)

            return 123

        self.agency.spy_on(func)

        d = func(42)
        self.assertEqual(d, 123)
        self.assertTrue(func.called)

    def test_call_with_bound_method_with_list_comprehension_and_self(self):
        """Testing FunctionSpy calls for bound method using a list
        comprehension referencing 'self'
        """
        obj = AdderObject()
        self.agency.spy_on(obj.func)

        result = obj.func()
        self.assertTrue(obj.func.called)
        self.assertEqual(result, [2, 3, 4])

    def test_call_with_unbound_method_with_list_comprehension_and_self(self):
        """Testing FunctionSpy calls for unbound method using a list
        comprehension referencing 'self'
        """
        self.agency.spy_on(AdderObject.func)

        obj = AdderObject()
        result = obj.func()
        self.assertTrue(obj.func.called)
        self.assertEqual(result, [2, 3, 4])

    def test_call_with_classmethod_with_list_comprehension_and_self(self):
        """Testing FunctionSpy calls for classmethod using a list
        comprehension referencing 'cls'
        """
        self.agency.spy_on(AdderObject.class_func)

        result = AdderObject.class_func()
        self.assertTrue(AdderObject.class_func.called)
        self.assertEqual(result, [2, 3, 4])

    def test_call_original_with_orig_func_and_function(self):
        """Testing FunctionSpy.call_original with spy on function set up using
        call_original=True
        """
        self.agency.spy_on(do_math,
                           call_original=True)

        result = do_math.call_original(10, 20)
        self.assertEqual(result, -10)

        self.assertFalse(do_math.called)

    def test_call_original_with_orig_func_and_bound_method(self):
        """Testing FunctionSpy.call_original with spy on bound method set up
        using call_original=True
        """
        obj = AdderObject()
        self.agency.spy_on(obj.add_one,
                           call_original=True)

        result = obj.add_one.call_original(5)
        self.assertEqual(result, 6)

        self.assertFalse(obj.add_one.called)

    def test_call_original_with_orig_func_and_unbound_method(self):
        """Testing FunctionSpy.call_original with spy on unbound method set up
        using call_original=True
        """
        self.agency.spy_on(AdderObject.add_one,
                           call_original=True)

        obj = AdderObject()
        result = obj.add_one.call_original(obj, 5)
        self.assertEqual(result, 6)

        self.assertFalse(AdderObject.add_one.called)
        self.assertFalse(obj.add_one.called)

    def test_call_original_with_orig_func_and_classmethod(self):
        """Testing FunctionSpy.call_original with spy on classmethod set up
        using call_original=True
        """
        self.agency.spy_on(AdderObject.class_add_one,
                           call_original=True)

        result = AdderObject.class_add_one.call_original(5)
        self.assertEqual(result, 6)

        self.assertFalse(AdderObject.class_add_one.called)

    def test_call_original_with_no_func_and_function(self):
        """Testing FunctionSpy.call_original with spy on function set up using
        call_original=False
        """
        self.agency.spy_on(do_math)

        result = do_math.call_original(10, 20)
        self.assertEqual(result, -10)

        self.assertFalse(do_math.called)

    def test_call_original_with_no_func_and_bound_method(self):
        """Testing FunctionSpy.call_original with spy on bound method set up
        using call_original=False
        """
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        result = obj.do_math.call_original(10, 20)
        self.assertEqual(result, 30)

        self.assertFalse(obj.do_math.called)

    def test_call_original_with_no_func_and_unbound_method(self):
        """Testing FunctionSpy.call_original with spy on unbound method set up
        using call_original=False
        """
        self.agency.spy_on(MathClass.do_math)

        obj = MathClass()
        result = obj.do_math.call_original(obj, 10, 20)
        self.assertEqual(result, 30)

        self.assertFalse(MathClass.do_math.called)
        self.assertFalse(obj.do_math.called)

    def test_call_original_with_no_func_and_classmethod(self):
        """Testing FunctionSpy.call_original with spy on classmethod set up
        using call_original=False
        """
        self.agency.spy_on(MathClass.class_do_math)

        result = MathClass.class_do_math.call_original(10, 20)
        self.assertEqual(result, 200)

        self.assertFalse(MathClass.class_do_math.called)

    def test_call_original_with_fake_func_and_function(self):
        """Testing FunctionSpy.call_original with spy on function set up using
        call_fake=
        """
        self.agency.spy_on(do_math,
                           call_fake=fake_do_math)

        result = do_math.call_original(10, 20)
        self.assertEqual(result, -10)

        self.assertFalse(do_math.called)

    def test_call_original_with_fake_func_and_bound_method(self):
        """Testing FunctionSpy.call_original with spy on bound method set up
        using call_fake=
        """
        obj = MathClass()
        self.agency.spy_on(obj.do_math,
                           call_fake=fake_do_math)

        result = obj.do_math.call_original(10, 20)
        self.assertEqual(result, 30)

        self.assertFalse(obj.do_math.called)

    def test_call_original_with_fake_func_and_unbound_method(self):
        """Testing FunctionSpy.call_original with spy on unbound method set up
        using call_fake=
        """
        self.agency.spy_on(MathClass.do_math,
                           call_fake=fake_do_math)

        obj = MathClass()
        result = obj.do_math.call_original(obj, 10, 20)
        self.assertEqual(result, 30)

        self.assertFalse(MathClass.do_math.called)
        self.assertFalse(obj.do_math.called)

    def test_call_original_with_fake_func_and_classmethod(self):
        """Testing FunctionSpy.call_original with spy on classmethod set up
        using call_fake=
        """
        self.agency.spy_on(MathClass.class_do_math,
                           call_fake=fake_class_do_math)

        result = MathClass.class_do_math.call_original(10, 20)
        self.assertEqual(result, 200)

        self.assertFalse(MathClass.class_do_math.called)

    def test_call_original_with_unbound_method_no_instance(self):
        """Testing FunctionSpy.call_original with spy on unbound method set up
        using call_original=True without passing instance
        """
        self.agency.spy_on(AdderObject.add_one,
                           call_original=True)

        obj = AdderObject()

        message = re.escape(
            'The first argument to add_one.call_original() must be an '
            'instance of kgb.tests.test_function_spy.AdderObject, since this '
            'is an unbound method.'
        )

        with self.assertRaisesRegex(TypeError, message):
            obj.add_one.call_original(5)

    def test_called(self):
        """Testing FunctionSpy.called"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        self.assertFalse(obj.do_math.called)

        obj.do_math(10, 20)

        self.assertTrue(obj.do_math.called)

    def test_last_call(self):
        """Testing FunctionSpy.last_call"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        obj.do_math(10, 20)
        obj.do_math(20, 30)

        self.assertEqual(len(obj.do_math.calls), 2)

        last_call = obj.do_math.last_call
        self.assertNotEqual(last_call, None)
        self.assertTrue(last_call.called_with(a=20, b=30))

    def test_last_call_with_no_calls(self):
        """Testing FunctionSpy.last_call on uncalled function"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        self.assertEqual(len(obj.do_math.calls), 0)

        last_call = obj.do_math.last_call
        self.assertEqual(last_call, None)

    def test_unspy(self):
        """Testing FunctionSpy.unspy"""
        orig_code = getattr(something_awesome, FunctionSig.FUNC_CODE_ATTR)
        spy = self.agency.spy_on(something_awesome, call_fake=lambda: 'spy!')

        self.assertTrue(hasattr(something_awesome, 'spy'))
        self.assertEqual(something_awesome.spy, spy)
        self.assertEqual(something_awesome(), 'spy!')

        spy.unspy()
        self.assertFalse(hasattr(something_awesome, 'spy'))
        self.assertEqual(getattr(something_awesome,
                                 FunctionSig.FUNC_CODE_ATTR),
                         orig_code)
        self.assertEqual(something_awesome(), 'Tada!')

    def test_unspy_and_bound_method(self):
        """Testing FunctionSpy.unspy and bound method"""
        obj = MathClass()
        func_dict = obj.do_math.__dict__.copy()

        spy = self.agency.spy_on(obj.do_math)
        self.assertNotEqual(obj.do_math.__dict__, func_dict)

        spy.unspy()
        self.assertEqual(obj.do_math.__dict__, func_dict)

    def test_unspy_with_bound_method_and_custom_setattr(self):
        """Testing FunctionSpy.unspy with a bound method on a class containing
        a custom __setattr__
        """
        class MyObject(object):
            def __setattr__(self, key, value):
                assert False

            def foo(self):
                pass

        obj = MyObject()
        func_dict = obj.foo.__dict__.copy()

        spy = self.agency.spy_on(obj.foo)
        self.assertNotEqual(obj.foo.__dict__, func_dict)

        spy.unspy()
        self.assertEqual(obj.foo.__dict__, func_dict)

    def test_unspy_and_unbound_method(self):
        """Testing FunctionSpy.unspy and unbound method"""
        func_dict = MathClass.do_math.__dict__.copy()

        spy = self.agency.spy_on(MathClass.do_math)
        self.assertNotEqual(MathClass.do_math.__dict__, func_dict)

        spy.unspy()
        self.assertEqual(MathClass.do_math.__dict__, func_dict)

    def test_unspy_with_classmethod(self):
        """Testing FunctionSpy.unspy with classmethod"""
        func_dict = MathClass.class_do_math.__dict__.copy()

        spy = self.agency.spy_on(MathClass.class_do_math)
        self.assertNotEqual(MathClass.class_do_math.__dict__, func_dict)

        spy.unspy()
        self.assertEqual(MathClass.class_do_math.__dict__, func_dict)

    def test_unspy_with_classmethod_on_parent(self):
        """Testing FunctionSpy.unspy with classmethod on parent class"""
        class MyParent(object):
            @classmethod
            def foo(self):
                pass

        class MyObject(MyParent):
            pass

        parent_func_dict = MyParent.foo.__dict__.copy()
        obj_func_dict = MyObject.foo.__dict__.copy()

        spy = self.agency.spy_on(MyObject.foo)
        self.assertNotEqual(MyObject.foo.__dict__, obj_func_dict)
        self.assertEqual(MyParent.foo.__dict__, parent_func_dict)

        spy.unspy()
        self.assertEqual(MyObject.foo.__dict__, obj_func_dict)
        self.assertEqual(MyParent.foo.__dict__, parent_func_dict)

    def test_unspy_with_unbound_method_on_parent(self):
        """Testing FunctionSpy.unspy with unbound method on parent class"""
        class MyParent(object):
            def foo(self):
                pass

        class MyObject(MyParent):
            pass

        parent_func_dict = MyParent.foo.__dict__.copy()
        obj_func_dict = MyObject.foo.__dict__.copy()

        spy = self.agency.spy_on(MyObject.foo, owner=MyObject)
        self.assertNotEqual(MyObject.foo.__dict__, obj_func_dict)
        self.assertEqual(MyParent.foo.__dict__, parent_func_dict)

        spy.unspy()
        self.assertEqual(MyObject.foo.__dict__, obj_func_dict)
        self.assertEqual(MyParent.foo.__dict__, parent_func_dict)

    def test_unspy_with_slippery_bound_method(self):
        """Testing FunctionSpy.unspy with slippery function generated by
        spied-on bound method
        """
        obj = SlipperyFuncObject()
        spy = self.agency.spy_on(obj.my_func, owner=obj)
        spy.unspy()

        self.assertFalse(hasattr(obj.my_func, 'spy'))
        self.assertNotIn('my_func', obj.__dict__)

        # Make sure the old behavior has reverted.
        slippery_func.count = 0

        func1 = obj.my_func
        func2 = obj.my_func
        self.assertIsNot(func1, func2)

        # These will have already had their values set, so we should get
        # stable results again.
        self.assertEqual(func1(), 0)
        self.assertEqual(func1(), 0)
        self.assertEqual(func2(), 1)
        self.assertEqual(func2(), 1)

        # These will trigger new counter increments when re-generating the
        # function in the decorator.
        self.assertEqual(obj.my_func(), 2)
        self.assertEqual(obj.my_func(), 3)

    def test_called_with(self):
        """Testing FunctionSpy.called_with"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(1, b=2)
        obj.do_math_mixed(3, b=4)

        self.assertTrue(obj.do_math_mixed.called_with(1, b=2))
        self.assertTrue(obj.do_math_mixed.called_with(3, b=4))
        self.assertTrue(obj.do_math_mixed.called_with(a=1, b=2))
        self.assertTrue(obj.do_math_mixed.called_with(a=3, b=4))
        self.assertFalse(obj.do_math_mixed.called_with(1, 2))
        self.assertFalse(obj.do_math_mixed.called_with(3, 4))
        self.assertFalse(obj.do_math_mixed.called_with(5, b=6))

    def test_called_with_and_keyword_args(self):
        """Testing FunctionSpy.called_with and keyword arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(a=1, b=2)
        obj.do_math_mixed(a=3, b=4)

        self.assertTrue(obj.do_math_mixed.called_with(1, b=2))
        self.assertTrue(obj.do_math_mixed.called_with(3, b=4))
        self.assertTrue(obj.do_math_mixed.called_with(a=1, b=2))
        self.assertTrue(obj.do_math_mixed.called_with(a=3, b=4))
        self.assertFalse(obj.do_math_mixed.called_with(5, b=6))

    def test_called_with_and_partial_args(self):
        """Testing FunctionSpy.called_with and partial arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(1, 2)
        obj.do_math_mixed(3, 4)

        self.assertTrue(obj.do_math_mixed.called_with(1))
        self.assertTrue(obj.do_math_mixed.called_with(3))
        self.assertFalse(obj.do_math_mixed.called_with(4))
        self.assertFalse(obj.do_math_mixed.called_with(1, 2, 3))

    def test_called_with_and_partial_kwargs(self):
        """Testing FunctionSpy.called_with and partial keyword arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(a=1, b=2)
        obj.do_math_mixed(a=3, b=4)

        self.assertTrue(obj.do_math_mixed.called_with(1))
        self.assertTrue(obj.do_math_mixed.called_with(b=2))
        self.assertTrue(obj.do_math_mixed.called_with(3))
        self.assertTrue(obj.do_math_mixed.called_with(b=4))
        self.assertTrue(obj.do_math_mixed.called_with(a=1, b=2))
        self.assertTrue(obj.do_math_mixed.called_with(a=3, b=4))
        self.assertFalse(obj.do_math_mixed.called_with(1, 2))
        self.assertFalse(obj.do_math_mixed.called_with(3, 4))
        self.assertFalse(obj.do_math_mixed.called_with(a=4))
        self.assertFalse(obj.do_math_mixed.called_with(a=1, b=2, c=3))
        self.assertFalse(obj.do_math_mixed.called_with(a=1, b=4))
        self.assertFalse(obj.do_math_mixed.called_with(a=3, b=2))

    def test_last_called_with(self):
        """Testing FunctionSpy.last_called_with"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(1, 2)
        obj.do_math_mixed(3, 4)

        self.assertFalse(obj.do_math_mixed.last_called_with(1, a=2))
        self.assertTrue(obj.do_math_mixed.last_called_with(3, b=4))

    def test_last_called_with_and_keyword_args(self):
        """Testing FunctionSpy.last_called_with and keyword arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(a=1, b=2)
        obj.do_math_mixed(a=3, b=4)

        self.assertTrue(obj.do_math_mixed.last_called_with(3, b=4))
        self.assertFalse(obj.do_math_mixed.last_called_with(1, b=2))
        self.assertFalse(obj.do_math_mixed.last_called_with(1, b=2, c=3))

    def test_last_called_with_and_partial_args(self):
        """Testing FunctionSpy.called_with and partial arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(1, 2)
        obj.do_math_mixed(3, 4)

        self.assertTrue(obj.do_math_mixed.last_called_with(3))
        self.assertTrue(obj.do_math_mixed.last_called_with(3, b=4))
        self.assertFalse(obj.do_math_mixed.last_called_with(3, b=4, c=5))
        self.assertFalse(obj.do_math_mixed.last_called_with(1, b=2))

    def test_last_called_with_and_partial_kwargs(self):
        """Testing FunctionSpy.called_with and partial keyword arguments"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math_mixed)

        obj.do_math_mixed(a=1, b=2)
        obj.do_math_mixed(a=3, b=4)

        self.assertTrue(obj.do_math_mixed.last_called_with(3))
        self.assertTrue(obj.do_math_mixed.last_called_with(b=4))
        self.assertFalse(obj.do_math_mixed.last_called_with(a=1))
        self.assertFalse(obj.do_math_mixed.last_called_with(b=2))
        self.assertFalse(obj.do_math_mixed.last_called_with(b=3))
        self.assertFalse(obj.do_math_mixed.last_called_with(3, 4))
        self.assertFalse(obj.do_math_mixed.last_called_with(a=1, b=2, c=3))
        self.assertFalse(obj.do_math_mixed.last_called_with(a=1, c=3))

    def test_returned(self):
        """Testing FunctionSpy.returned"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        obj.do_math(1, 2)
        obj.do_math(3, 4)

        self.assertTrue(obj.do_math.returned(3))
        self.assertTrue(obj.do_math.returned(7))
        self.assertFalse(obj.do_math.returned(10))
        self.assertFalse(obj.do_math.returned(None))

    def test_last_returned(self):
        """Testing FunctionSpy.last_returned"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        obj.do_math(1, 2)
        obj.do_math(3, 4)

        self.assertFalse(obj.do_math.last_returned(3))
        self.assertTrue(obj.do_math.last_returned(7))
        self.assertFalse(obj.do_math.last_returned(None))

    def test_raised(self):
        """Testing FunctionSpy.raised"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        with self.assertRaises(TypeError):
            obj.do_math(1, 'a')

        self.assertTrue(obj.do_math.raised(TypeError))
        self.assertFalse(obj.do_math.raised(ValueError))
        self.assertFalse(obj.do_math.raised(None))

    def test_last_raised(self):
        """Testing FunctionSpy.last_raised"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        with self.assertRaises(TypeError):
            obj.do_math(1, 'a')

        self.assertTrue(obj.do_math.last_raised(TypeError))
        self.assertFalse(obj.do_math.last_raised(None))

        obj.do_math(1, 4)

        self.assertFalse(obj.do_math.last_raised(TypeError))
        self.assertTrue(obj.do_math.last_raised(None))

    def test_raised_with_message(self):
        """Testing FunctionSpy.raised_with_message"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        with self.assertRaises(TypeError):
            obj.do_math(1, 'a')

        self.assertTrue(obj.do_math.raised_with_message(
            TypeError,
            "unsupported operand type(s) for +: 'int' and '%s'"
            % text_type.__name__))
        self.assertFalse(obj.do_math.raised_with_message(
            ValueError,
            "unsupported operand type(s) for +: 'int' and '%s'"
            % text_type.__name__))
        self.assertFalse(obj.do_math.raised_with_message(TypeError, None))

    def test_last_raised_with_message(self):
        """Testing FunctionSpy.last_raised_with_message"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        with self.assertRaises(TypeError):
            obj.do_math(1, 'a')

        self.assertTrue(obj.do_math.last_raised_with_message(
            TypeError,
            "unsupported operand type(s) for +: 'int' and '%s'"
            % text_type.__name__))
        self.assertFalse(obj.do_math.last_raised_with_message(TypeError, None))

    def test_reset_calls(self):
        """Testing FunctionSpy.reset_calls"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        obj.do_math(1, 2)
        self.assertEqual(len(obj.do_math.calls), 1)
        self.assertEqual(obj.do_math.last_call, obj.do_math.calls[-1])
        self.assertTrue(obj.do_math.called)

        obj.do_math.reset_calls()
        self.assertEqual(len(obj.do_math.calls), 0)
        self.assertIsNone(obj.do_math.last_call)
        self.assertFalse(obj.do_math.called)

    def test_repr(self):
        """Testing FunctionSpy.__repr__"""
        self.agency.spy_on(something_awesome)

        self.assertTrue(hasattr(something_awesome, 'spy'))
        self.assertTrue(repr(something_awesome.spy),
                        '<Spy for something_awesome (0 calls)>')

    def test_repr_and_function(self):
        """Testing FunctionSpy.__repr__ and function"""
        self.agency.spy_on(do_math)

        self.assertEqual(repr(do_math.spy),
                         '<Spy for function do_math (0 calls)>')

    def test_repr_and_bound_method(self):
        """Testing FunctionSpy.__repr__ and bound method"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        obj.do_math()

        self.assertEqual(repr(obj.do_math.spy),
                         '<Spy for bound method MathClass.do_math '
                         'of %r (1 call)>' % obj)

    def test_repr_and_unbound_method(self):
        """Testing FunctionSpy.__repr__ and unbound method"""
        self.agency.spy_on(MathClass.do_math)

        self.assertEqual(repr(MathClass.do_math.spy),
                         '<Spy for unbound method MathClass.do_math '
                         'of %r (0 calls)>' % MathClass)

    def test_repr_with_classmethod(self):
        """Testing FunctionSpy.__repr__ with classmethod"""
        self.agency.spy_on(MathClass.class_do_math)

        self.assertEqual(
            repr(MathClass.class_do_math.spy),
            '<Spy for classmethod MathClass.class_do_math of %r (0 calls)>'
            % MathClass)

    @require_getargspec
    def test_getargspec_with_function(self):
        """Testing FunctionSpy in inspect.getargspec() with function"""
        self.agency.spy_on(do_math)

        args, varargs, keywords, defaults = inspect.getargspec(do_math)
        self.assertEqual(args, ['a', 'b'])
        self.assertEqual(varargs, 'args')
        self.assertEqual(keywords, 'kwargs')
        self.assertEqual(defaults, (1, 2))

    @require_getargspec
    def test_getargspec_with_bound_method(self):
        """Testing FunctionSpy in inspect.getargspec() with bound method"""
        obj = MathClass()
        self.agency.spy_on(obj.do_math)

        args, varargs, keywords, defaults = inspect.getargspec(obj.do_math)
        self.assertEqual(args, ['self', 'a', 'b'])
        self.assertEqual(varargs, 'args')
        self.assertEqual(keywords, 'kwargs')
        self.assertEqual(defaults, (1, 2))

    @require_getargspec
    def test_getargspec_with_unbound_method(self):
        """Testing FunctionSpy in inspect.getargspec() with unbound method"""
        self.agency.spy_on(MathClass.do_math)

        args, varargs, keywords, defaults = \
            inspect.getargspec(MathClass.do_math)
        self.assertEqual(args, ['self', 'a', 'b'])
        self.assertEqual(varargs, 'args')
        self.assertEqual(keywords, 'kwargs')
        self.assertEqual(defaults, (1, 2))

    @require_getargspec
    def test_getargspec_with_classmethod(self):
        """Testing FunctionSpy in inspect.getargspec() with classmethod"""
        obj = MathClass()
        self.agency.spy_on(obj.class_do_math)

        args, varargs, keywords, defaults = \
            inspect.getargspec(obj.class_do_math)
        self.assertEqual(args, ['cls', 'a', 'b'])
        self.assertEqual(varargs, 'args')
        self.assertEqual(keywords, 'kwargs')
        self.assertEqual(defaults, (2, 5))

    def test_spy_on_generator(self) -> None:
        """Testing FunctionSpy with a generator function"""
        # Python 3.13+ added a deprecation warning when replacing a generator
        # function. This test verifies that our fix for this does not trigger a
        # DeprecationWarning. See the comments in FunctionSpy._build_spy_code
        # for details on the issue.
        def _generator_function():
            yield 'a'
            yield 'b'
            yield 'c'

        def _fake_generator():
            yield 'x'
            yield 'y'
            yield 'z'

        with warnings.catch_warnings():
            warnings.simplefilter('error', DeprecationWarning)

            spy = self.agency.spy_on(_generator_function,
                                     call_fake=_fake_generator)
            results = list(_generator_function())
            self.assertEqual(results, ['x', 'y', 'z'])
            spy.unspy()

            spy = self.agency.spy_on(_generator_function)

            results = list(_generator_function())
            self.assertEqual(results, ['a', 'b', 'c'])
            spy.unspy()



"""Spy-related errors."""

from __future__ import unicode_literals

import traceback


class InternalKGBError(Exception):
    """An internal error about the inner workings of KGB."""

    def __init__(self, msg):
        """Initialize the error.

        Args:
            msg (unicode):
                The message to display. A general message about contacting
                support will be appended to this.
        """
        super(InternalKGBError, self).__init__(
            '%s\n\n'
            'This is an internal error in KGB. Please report it!'
            % msg)


class ExistingSpyError(ValueError):
    """An error for when an existing spy was found on a function.

    This will provide a helpful error message explaining what went wrong,
    showing a backtrace of the original spy's setup in order to help diagnose
    the problem.
    """

    def __init__(self, func):
        """Initialize the error.

        Args:
            func (callable):
                The function containing an existing spy.
        """
        super(ExistingSpyError, self).__init__(
            'The function %(func)r has already been spied on. Here is where '
            'that spy was set up:\n\n'
            '%(stacktrace)s\n'
            'You may have encountered a crash in that test preventing the '
            'spy from being unregistered. Try running that test manually.'
            % {
                'func': func,
                'stacktrace': ''.join(traceback.format_stack(
                    func.spy.init_frame.f_back)[-4:]),
            })


class IncompatibleFunctionError(ValueError):
    """An error for when a function signature is incompatible.

    This is used for the ``call_fake`` function passed in when setting up a
    spy.
    """

    def __init__(self, func, func_sig, incompatible_func,
                 incompatible_func_sig):
        """Initialize the error.

        Args:
            func (callable):
                The function containing the original signature.

            func_sig (kgb.signature.FunctionSig):
                The signature of ``func``.

            incompatible_func (callable):
                The function that was not compatible.

            incompatible_func_sig (kgb.signature.FunctionSig):
                The signature of ``incompatible_func``.
        """
        super(IncompatibleFunctionError, self).__init__(
            'The function signature of %r (%s) is not compatible with %r (%s).'
            % (incompatible_func,
               incompatible_func_sig.format_arg_spec(),
               func,
               func_sig.format_arg_spec()))


class UnexpectedCallError(AssertionError):
    """A call was made to a spy that was not expected."""



"""Independent assertion functions.

These assertion functions can be used in pytest or in other places where
:py:class:`kgb.SpyAgency` can't be mixed.

Version Added:
    7.0
"""

from __future__ import unicode_literals

from kgb.agency import SpyAgency


_agency = SpyAgency()

__all__ = []


for name in vars(SpyAgency):
    if name.startswith('assert_'):
        globals()[name] = getattr(_agency, name)
        __all__.append(name)



"""Function signature introspection and code generation."""

from __future__ import unicode_literals

import inspect
import logging
import sys
import types

from kgb.errors import InternalKGBError
from kgb.utils import get_defined_attr_value


logger = logging.getLogger('kgb')


class _UnsetArg(object):
    """Internal class for representation unset arguments on functions."""

    def __repr__(self):
        """Return a string representation of this object.

        Returns:
            unicode:
            ``_UNSET_ARG``.
        """
        return '_UNSET_ARG'


_UNSET_ARG = _UnsetArg()


class BaseFunctionSig(object):
    """Base class for a function signature introspector.

    This is responsible for taking a function (and a user-requested owner)
    and determining the actual owner, function type, function name, and
    arguments. It's also responsible for generating code that can be used
    to help define functions or perform calls, for use in spy code generation.

    How this is all done depends entirely on the version of Python. Subclasses
    must implement all this logic.

    Version Changed:
        5.0:
        Added support for the following attributes:

        * :py:attr:`defined_func`.
        * :py:attr:`has_getter`.
        * :py:attr:`has_setter`.
        * :py:attr:`is_slippery`.

    Attributes:
        defined_func (callable or object):
            The actual function (or wrapping object) that's defined on
            somewhere in the owner's class hierarchy (or the function itself if
            this is a standalone function). This may differ from
            :py:attr:`func`.

        has_getter (bool):
            Whether this signature represents a descriptor with a ``__get__``
            method.

        has_setter (bool):
            Whether this signature represents a descriptor with a ``__set__``
            method.

        is_slippery (bool):
            Whether this represents a slippery function. This is a method on
            a class that returns a different function every time its attribute
            is accessed on an instance.

            This occurs when a method decorator is used that wraps a function
            on access and returns the wrapper function, but does not cache the
            wrapper function. These are returned as standard functions and
            not methods.

            Slippery functions can only be detected when an explicit owner is
            provided.
    """

    #: The signature represents a standard function.
    TYPE_FUNCTION = 0

    #: The signature represents a bound method.
    #:
    #: Bound methods are functions on an instance of a class, or classmethods.
    TYPE_BOUND_METHOD = 1

    #: The signature represents an unbound method.
    #:
    #: Unbound methods are standard methods on a class.
    TYPE_UNBOUND_METHOD = 2

    def __init__(self, func, owner=_UNSET_ARG, func_name=None):
        """Initialize the signature.

        Subclasses must override this to parse function types/ownership and
        available arguments. They must call :py:meth:`finalize_state` once
        they've calculated all signature state.

        Args:
            func (callable):
                The function to use for the signature.

            owner (type or object, optional):
                The owning class, as provided when spying on the function.
                This is not stored directly (as it may be invalid), but can
                be used for informative purposes for subclasses.

            func_name (str, optional):
                An explicit name for the function. This will be used instead
                of the function's specified name, and is usually a sign of a
                bad decorator.

                Version Added:
                    7.0
        """
        self.func = func
        self.func_type = self.TYPE_FUNCTION
        self.func_name = func_name or getattr(func, self.FUNC_NAME_ATTR)
        self.owner = None

        if hasattr(func, '__func__'):
            # This is an instancemethod on a class. Grab the real function
            # from it.
            self.real_func = func.__func__
        else:
            self.real_func = func

        self.all_arg_names = []
        self.arg_names = []
        self.kwarg_names = []
        self.args_param_name = []
        self.kwargs_param_name = []
        self.is_slippery = False
        self.has_getter = False
        self.has_setter = False

    def is_compatible_with(self, other_sig):
        """Return whether two function signatures are compatible.

        This will check if the signature for a function (the ``call_fake``
        passed in, technically) is compatible with another (the spied
        function), to help ensure that unit tests with incompatible function
        signatures don't blow up with strange errors later.

        This will attempt to be somewhat flexible in what it considers
        compatible. Basically, so long as all the arguments passed in to
        the source function could be resolved using the argument list in the
        other function (taking into account things like positional argument
        names as keyword arguments), they're considered compatible.

        Args:
            other_sig (BaseFunctionSig):
                The other signature to check for compatibility with.

        Returns:
            bool:
            ``True`` if ``other_sig`` is considered compatible with this
            signature. ``False`` if it is not.
        """
        source_args_name = self.args_param_name
        compat_args_name = other_sig.args_param_name
        source_kwargs_name = self.kwargs_param_name
        compat_kwargs_name = other_sig.kwargs_param_name

        if compat_args_name and compat_kwargs_name:
            return True

        if ((source_args_name and not compat_args_name) or
            (source_kwargs_name and not compat_kwargs_name)):
            return False

        source_args = self.arg_names
        compat_args = other_sig.arg_names
        compat_all_args = set(other_sig.all_arg_names)
        compat_kwargs = set(other_sig.kwarg_names)

        if self.func_type in (self.TYPE_BOUND_METHOD,
                              self.TYPE_UNBOUND_METHOD):
            source_args = source_args[1:]
            compat_args = compat_args[1:]

        if (len(source_args) != len(compat_args) and
            ((len(source_args) < len(compat_args) and not source_args_name and
              not compat_kwargs.issuperset(source_args)) or
             (len(source_args) > len(compat_args) and not compat_args_name))):
            return False

        if (not compat_all_args.issuperset(self.kwarg_names) and
            not compat_kwargs_name):
            return False

        return True

    def format_forward_call_args(self):
        """Format arguments to pass in for forwarding a call.

        This will build a string for use in the forwarding call, which will
        pass every positional and keyword parameter defined for the function
        to forwarded function, along with the ``*args`` and ``**kwargs``,
        if specified.

        Returns:
            unicode:
            A string representing the arguments to pass when forwarding a call.
        """
        _format_arg = self.format_forward_call_arg

        # Build the list of positional and keyword arguments.
        result = [
            _format_arg(arg_name)
            for arg_name in self.arg_names
        ] + [
            '%s=%s' % (arg_name, _format_arg(arg_name))
            for arg_name in self.kwarg_names
        ]

        # Add the variable arguments.
        if self.args_param_name:
            result.append('*%s' % _format_arg(self.args_param_name))

        if self.kwargs_param_name:
            result.append('**%s' % _format_arg(self.kwargs_param_name))

        return ', '.join(result)

    def format_forward_call_arg(self, arg_name):
        """Return a string used to reference an argument in a forwarding call.

        Subclasses must implement this to return code the spy can use when
        generating a function to forward arguments in a call.

        Args:
            arg_name (unicode):
                The name of the argument.

        Returns:
            unicode:
            The string used to format the argument call.
        """
        raise NotImplementedError

    def format_arg_spec(self):
        """Format the function's arguments for a new function definition.

        This will build a list of parameters for a function definition based on
        the argument specification found when introspecting a spied function.
        This consists of all supported argument types for the version of
        Python.

        Returns:
            unicode:
            A string representing an argument list for a function definition.
        """
        raise NotImplementedError

    def finalize_state(self):
        """Finalize the state for the signature.

        This will set any remaining values for the signature based on the
        calculations already performed by the subclasses. This must be called
        at the end of a subclass's :py:meth:`__init__`.
        """
        if self.owner is None:
            self.defined_func = self.func
        else:
            try:
                self.defined_func = get_defined_attr_value(self.owner,
                                                           self.func_name)
            except AttributeError:
                # This was a dynamically-injected function. We won't find it
                # in the class hierarchy. Use the provided function instead.
                self.defined_func = self.func

        if not isinstance(self.defined_func, (types.FunctionType,
                                              types.MethodType,
                                              classmethod,
                                              staticmethod)):
            if hasattr(self.defined_func, '__get__'):
                self.has_getter = True

            if hasattr(self.defined_func, '__set__'):
                self.has_setter = True


class FunctionSigPy2(BaseFunctionSig):
    """Function signature introspector for Python 2.

    This supports introspecting functions and generating code for use in
    spies when running on Python 2.
    """

    FUNC_CLOSURE_ATTR = 'func_closure'
    FUNC_CODE_ATTR = 'func_code'
    FUNC_DEFAULTS_ATTR = 'func_defaults'
    FUNC_GLOBALS_ATTR = 'func_globals'
    FUNC_NAME_ATTR = 'func_name'
    METHOD_SELF_ATTR = 'im_self'

    def __init__(self, func, owner=_UNSET_ARG, func_name=None):
        """Initialize the signature.

        Subclasses must override this to parse function types/ownership and
        available arguments.

        Args:
            func (callable):
                The function to use for the signature.

            owner (type, optional):
                The owning class, as provided when spying on the function.
                The value is ignored for methods on which an owner can be
                calculated, favoring the calculated value instead.

            func_name (str, optional):
                An explicit name for the function. This will be used instead
                of the function's specified name, and is usually a sign of a
                bad decorator.

                Version Added:
                    7.0
        """
        super(FunctionSigPy2, self).__init__(func=func,
                                             owner=owner,
                                             func_name=func_name)

        func_name = self.func_name

        # Figure out the owner and method type.
        if inspect.ismethod(func):
            # This is a bound or unbound method. If it's unbound, and an
            # owner is not specified, we're going to need to warn the user,
            # since things are going to break on Python 3.
            #
            # Otherwise, we're going to determine the bound vs. unbound type
            # and use the owner specified by the method. (The provided owner
            # will be validated in FunctionSpy.)
            method_owner = func.im_self

            if method_owner is None:
                self.func_type = self.TYPE_UNBOUND_METHOD
                self.owner = func.im_class

                if owner is _UNSET_ARG:
                    logger.warning('Unbound method owners can easily be '
                                   'determined on Python 2.x, but not on '
                                   '3.x. Please pass owner= to spy_on() '
                                   'to set a specific owner for %r.',
                                   func)
            else:
                self.func_type = self.TYPE_BOUND_METHOD
                self.owner = method_owner
        elif owner is not _UNSET_ARG:
            # This is a standard function, but an owner (as an instance) has
            # been provided. Find out if the owner has this function (either
            # the actual instance or one with the same name and bytecode),
            # and if so, treat this as a bound method.
            #
            # This is necessary when trying to spy on a decorated method that
            # generates functions dynamically (using something like
            # functools.wraps). We call these "slippery functions." A
            # real-world example is something like Stripe's
            # stripe.Customer.delete() method, which is a different function
            # every time you call it.
            owner_func = getattr(owner, func_name, None)

            if (owner_func is not None and
                (owner_func is func or
                 owner_func.func_code is func.func_code)):
                if inspect.isclass(owner):
                    self.func_type = self.TYPE_UNBOUND_METHOD
                else:
                    self.func_type = self.TYPE_BOUND_METHOD

                self.owner = owner
                self.is_slippery = owner_func is not func

        # Load information on the arguments.
        argspec = inspect.getargspec(func)
        all_args = argspec.args
        defaults = argspec.defaults

        if all_args and defaults:
            num_defaults = len(defaults)
            keyword_args = all_args[-num_defaults:]
            pos_args = all_args[:-num_defaults]
        else:
            pos_args = all_args
            keyword_args = []

        self.all_arg_names = argspec.args
        self.arg_names = pos_args
        self.kwarg_names = keyword_args
        self.args_param_name = argspec.varargs
        self.kwargs_param_name = argspec.keywords
        self._defaults = argspec.defaults

        self.finalize_state()

    def format_forward_call_arg(self, arg_name):
        """Return a string used to reference an argument in a forwarding call.

        Args:
            arg_name (unicode):
                The name of the argument.

        Returns:
            unicode:
            The string used to format the argument call.
        """
        return arg_name

    def format_arg_spec(self):
        """Format the function's arguments for a new function definition.

        This will build a list of parameters for a function definition based on
        the argument specification found when introspecting a spied function.
        This consists of all positional arguments, keyword arguments, and the
        special ``*args`` and ``**kwargs`` arguments.

        Returns:
            unicode:
            A string representing an argument list for a function definition.
        """
        return inspect.formatargspec(
            args=self.all_arg_names,
            varargs=self.args_param_name,
            varkw=self.kwargs_param_name,
            defaults=self._defaults,
            formatvalue=lambda value: '=_UNSET_ARG')[1:-1]


class FunctionSigPy3(BaseFunctionSig):
    """Function signature introspector for Python 3.

    This supports introspecting functions and generating code for use in
    spies when running on Python 3.

    There are some differences in function capabilities between Python 3.x
    releases (such as the addition of positional-only keyword arguments).
    This class provides compatibility for all these versions, currently up
    through Python 3.8.
    """

    FUNC_CLOSURE_ATTR = '__closure__'
    FUNC_CODE_ATTR = '__code__'
    FUNC_DEFAULTS_ATTR = '__defaults__'
    FUNC_GLOBALS_ATTR = '__globals__'
    FUNC_NAME_ATTR = '__name__'
    METHOD_SELF_ATTR = '__self__'

    def __init__(self, func, owner=_UNSET_ARG, func_name=None):
        """Initialize the signature.

        Subclasses must override this to parse function types/ownership and
        available arguments.

        Args:
            func (callable):
                The function to use for the signature.

            owner (type, optional):
                The owning class, as provided when spying on the function.
                This is used only when spying on unbound or slippery methods.

            func_name (str, optional):
                An explicit name for the function. This will be used instead
                of the function's specified name, and is usually a sign of a
                bad decorator.

                Version Added:
                    7.0
        """
        super(FunctionSigPy3, self).__init__(func=func,
                                             owner=owner,
                                             func_name=func_name)

        if not hasattr(inspect, '_signature_from_callable'):
            raise InternalKGBError(
                'Python %s.%s does not have inspect._signature_from_callable, '
                'which is needed in order to generate a Signature from a '
                'function.'
                % sys.version_info[:2])

        func_name = self.func_name

        # Figure out the owner and method type.
        #
        # Python 3 does not officially have unbound methods. Methods on
        # instances are easily identified as types.MethodType, but
        # unbound methods are just standard functions without something
        # like __self__ to point to the parent class.
        #
        # However, the owner can generally be inferred (but not always!).
        # Python 3.3 introduced __qualname__, which is a string
        # identifying the path to the class within the containing module.
        # The path is expected to be traversable, unless it contains
        # "<locals>" in it, in which case it's defined somewhere you can't
        # get to it (like in a function).
        #
        # So to determine if it's an unbound method, we check to see what
        # __qualname__ looks like, and then we try to find it. If we can,
        # we grab the owner and identify it as an unbound method. If not,
        # it stays as a standard function.
        if inspect.ismethod(func):
            self.func_type = self.TYPE_BOUND_METHOD
            self.owner = func.__self__
        elif '.' in func.__qualname__:
            if owner is not _UNSET_ARG:
                self.owner = owner

                try:
                    self.is_slippery = (
                        owner is not _UNSET_ARG and
                        getattr(owner, func_name) is not func
                    )
                except AttributeError:
                    if '<locals>' in func.__qualname__:
                        logger.warning(
                            "%r doesn't have a function named \"%s\". This "
                            "appears to be a decorator that doesn't "
                            "preserve function names. Try passing "
                            "func_name= when setting up the spy.",
                            owner, func_name)
                    else:
                        logger.warning(
                            "%r doesn't have a function named \"%s\". It's "
                            "not clear why this is. Try passing func_name= "
                            "when setting up the spy.",
                            owner, func_name)

                if owner is _UNSET_ARG or inspect.isclass(owner):
                    self.func_type = self.TYPE_UNBOUND_METHOD
                else:
                    self.func_type = self.TYPE_BOUND_METHOD
            elif '<locals>' in func.__qualname__:
                # We can only assume this is a function. It might not be.
                self.func_type = self.TYPE_FUNCTION
            else:
                real_func = self.real_func
                method_owner = inspect.getmodule(real_func)

                for part in real_func.__qualname__.split('.')[:-1]:
                    try:
                        method_owner = getattr(method_owner, part)
                    except AttributeError:
                        method_owner = None
                        break

                if method_owner is not None:
                    self.func_type = self.TYPE_UNBOUND_METHOD
                    self.owner = method_owner

                logger.warning('Determined the owner of %r to be %r, '
                               'but it may be wrong. Please pass '
                               'owner= to spy_on() to set a specific '
                               'owner.',
                               func, self.owner)

        # Load information on the arguments.
        sig = inspect._signature_from_callable(
            func,
            follow_wrapper_chains=False,
            skip_bound_arg=False,
            sigcls=inspect.Signature)

        all_args = []
        args = []
        kwargs = []

        for param in sig.parameters.values():
            kind = param.kind
            name = param.name

            if kind is param.POSITIONAL_OR_KEYWORD:
                # Standard arguments -- either positional or keyword.
                all_args.append(name)

                if param.default is param.empty:
                    args.append(name)
                else:
                    kwargs.append(name)
            elif kind is param.POSITIONAL_ONLY:
                # Positional-only arguments (Python 3.8+).
                all_args.append(name)
                args.append(name)
            elif kind is param.KEYWORD_ONLY:
                # Keyword-only arguments (Python 3+).
                kwargs.append(name)
            elif kind is param.VAR_POSITIONAL:
                # *args
                self.args_param_name = name
            elif kind is param.VAR_KEYWORD:
                # **kwargs
                self.kwargs_param_name = name

        self.all_arg_names = all_args
        self.arg_names = args
        self.kwarg_names = kwargs
        self._sig = sig

        self.finalize_state()

    def format_forward_call_arg(self, arg_name):
        """Return a string used to reference an argument in a forwarding call.

        Args:
            arg_name (unicode):
                The name of the argument.

        Returns:
            unicode:
            The string used to format the argument call.
        """
        # Starting in Python 3, something changed with variables. Due to
        # the way we generate the hybrid code object, we can't always
        # reference the local variables directly. Sometimes we can, but
        # other times we have to get them from locals(). We can't always
        # get them from there, though, so instead we conditionally check
        # both. This is wordy, but necessary.
        return '_kgb_l["%(arg)s"] if "%(arg)s" in _kgb_l else %(arg)s' % {
            'arg': arg_name,
        }

    def format_arg_spec(self):
        """Format the function's arguments for a new function definition.

        This will build a list of parameters for a function definition based on
        the argument specification found when introspecting a spied function.
        This consists of all positional arguments, positional-only arguments,
        keyword arguments, keyword-only arguments, and the special ``*args``
        and ``**kwargs`` arguments.

        Returns:
            unicode:
            A string representing an argument list for a function definition.
        """
        parameters = []

        # Make a copy of the Signature and its parameters, but leave out
        # all type annotations.
        for orig_param in self._sig.parameters.values():
            default = orig_param.default

            if (orig_param.kind is orig_param.POSITIONAL_OR_KEYWORD and
                default is not orig_param.empty):
                default = _UNSET_ARG

            parameters.append(inspect.Parameter(
                name=orig_param.name,
                kind=orig_param.kind,
                default=default))

        sig = inspect.Signature(parameters=parameters)

        return str(sig)[1:-1]


if sys.version_info[0] == 2:
    FunctionSig = FunctionSigPy2
elif sys.version_info[0] == 3:
    FunctionSig = FunctionSigPy3
else:
    raise Exception('Unsupported Python version')
